import pytest
import numpy as np

from nwsspc.sharp.calc import interp
from nwsspc.sharp.calc import constants


def test_interp_height():
    hght = np.arange(100.0, 1100.0, 100.0)
    data = np.arange(1.0, 11.0, 1.0)

    # Test missing/nan/inf behavior
    assert (interp.interp_height(0, hght, data) == constants.MISSING)
    assert (interp.interp_height(1100, hght, data) == constants.MISSING)
    assert (interp.interp_height(constants.MISSING,
            hght, data) == constants.MISSING)
    assert (interp.interp_height(np.inf, hght, data) == constants.MISSING)
    assert (interp.interp_height(np.nan, hght, data) == constants.MISSING)

    # Test exact values along the edges of the arrays
    assert (interp.interp_height(100, hght, data) == 1)
    assert (interp.interp_height(1000, hght, data) == 10)

    # Test exact value in the middle
    assert (interp.interp_height(500, hght, data) == 5)

    # Test between levels
    assert (interp.interp_height(550, hght, data) == 5.5)
    assert (interp.interp_height(110, hght, data) == pytest.approx(1.1))
    assert (interp.interp_height(391, hght, data) == pytest.approx(3.91))


def test_interp_pres():
    pres = np.arange(10000.0, 110000.0, 10000.0)[::-1]
    data = np.arange(1.0, 11.0, 1.0)

    # Test missing/nan/inf behavior
    assert (interp.interp_pressure(0, pres, data) == constants.MISSING)
    assert (interp.interp_pressure(110000.0, pres, data) == constants.MISSING)
    assert (interp.interp_pressure(
        constants.MISSING, pres, data) == constants.MISSING)
    assert (interp.interp_pressure(np.inf, pres, data) == constants.MISSING)
    assert (interp.interp_pressure(np.nan, pres, data) == constants.MISSING)

    # Test exact values along the edges of the array
    assert (interp.interp_pressure(100000.0, pres, data) == 1)
    assert (interp.interp_pressure(10000.0, pres, data) == 10)

    # Test an exact value in the middle
    assert (interp.interp_pressure(50000.0, pres, data) == 6)

    # Test between levels
    assert (interp.interp_pressure(97500.0, pres, data)
            == pytest.approx(1.2402969255))
    assert (interp.interp_pressure(95000.0, pres, data)
            == pytest.approx(1.4868382))
    assert (interp.interp_pressure(92500.0, pres, data)
            == pytest.approx(1.73995423))


def test_find_first_pres():
    pres = np.arange(10000.0, 110000.0, 10000.0)[::-1]
    data = np.arange(1.0, 11.0, 1.0)
    data[2] = constants.MISSING

    assert (interp.find_first_pressure(5.0, pres, data) == 60000.0)
    assert (interp.find_first_pressure(
        5.5, pres, data) == pytest.approx(54772.26))
    assert (interp.find_first_pressure(5.0, pres, data[::-1]) == 50000.0)
    assert (interp.find_first_pressure(
        5.5, pres, data[::-1]) == pytest.approx(54772.26))

    assert (interp.find_first_pressure(
        constants.MISSING, pres, data) == constants.MISSING)
    assert (interp.find_first_pressure(
        3, pres, data) == pytest.approx(79372.6))


def test_find_first_height():
    hght = np.arange(100.0, 1100.0, 100.0)
    data = np.arange(1.0, 11.0, 1.0)
    data[2] = constants.MISSING

    assert (interp.find_first_height(5.0, hght, data) == 500.0)
    assert (interp.find_first_height(5.5, hght, data) == 550.0)
    assert (interp.find_first_height(5.0, hght, data[::-1]) == 600.0)
    assert (interp.find_first_height(5.5, hght, data[::-1]) == 550.0)

    assert (interp.find_first_height(
        constants.MISSING, hght, data) == constants.MISSING)
    assert (interp.find_first_height(3, hght, data) == 300.0)



# type: ignore

"""
report.py - Utilities for reporting statistics about benchmark results
"""

import copy
import os
import random
import re
import unittest

from numpy import array
from scipy.stats import gmean, mannwhitneyu


class BenchmarkColor(object):
    def __init__(self, name, code):
        self.name = name
        self.code = code

    def __repr__(self):
        return "%s%r" % (self.__class__.__name__, (self.name, self.code))

    def __format__(self, format):
        return self.code


# Benchmark Colors Enumeration
BC_NONE = BenchmarkColor("NONE", "")
BC_MAGENTA = BenchmarkColor("MAGENTA", "\033[95m")
BC_CYAN = BenchmarkColor("CYAN", "\033[96m")
BC_OKBLUE = BenchmarkColor("OKBLUE", "\033[94m")
BC_OKGREEN = BenchmarkColor("OKGREEN", "\033[32m")
BC_HEADER = BenchmarkColor("HEADER", "\033[92m")
BC_WARNING = BenchmarkColor("WARNING", "\033[93m")
BC_WHITE = BenchmarkColor("WHITE", "\033[97m")
BC_FAIL = BenchmarkColor("FAIL", "\033[91m")
BC_ENDC = BenchmarkColor("ENDC", "\033[0m")
BC_BOLD = BenchmarkColor("BOLD", "\033[1m")
BC_UNDERLINE = BenchmarkColor("UNDERLINE", "\033[4m")

UTEST_MIN_REPETITIONS = 2
UTEST_OPTIMAL_REPETITIONS = 9  # Lowest reasonable number, More is better.
UTEST_COL_NAME = "_pvalue"

_TIME_UNIT_TO_SECONDS_MULTIPLIER = {
    "s": 1.0,
    "ms": 1e-3,
    "us": 1e-6,
    "ns": 1e-9,
}


def color_format(use_color, fmt_str, *args, **kwargs):
    """
    Return the result of 'fmt_str.format(*args, **kwargs)' after transforming
    'args' and 'kwargs' according to the value of 'use_color'. If 'use_color'
    is False then all color codes in 'args' and 'kwargs' are replaced with
    the empty string.
    """
    assert use_color is True or use_color is False
    if not use_color:
        args = [
            arg if not isinstance(arg, BenchmarkColor) else BC_NONE
            for arg in args
        ]
        kwargs = {
            key: arg if not isinstance(arg, BenchmarkColor) else BC_NONE
            for key, arg in kwargs.items()
        }
    return fmt_str.format(*args, **kwargs)


def find_longest_name(benchmark_list):
    """
    Return the length of the longest benchmark name in a given list of
    benchmark JSON objects
    """
    longest_name = 1
    for bc in benchmark_list:
        if len(bc["name"]) > longest_name:
            longest_name = len(bc["name"])
    return longest_name


def calculate_change(old_val, new_val):
    """
    Return a float representing the decimal change between old_val and new_val.
    """
    if old_val == 0 and new_val == 0:
        return 0.0
    if old_val == 0:
        return float(new_val - old_val) / (float(old_val + new_val) / 2)
    return float(new_val - old_val) / abs(old_val)


def filter_benchmark(json_orig, family, replacement=""):
    """
    Apply a filter to the json, and only leave the 'family' of benchmarks.
    """
    regex = re.compile(family)
    filtered = {}
    filtered["benchmarks"] = []
    for be in json_orig["benchmarks"]:
        if not regex.search(be["name"]):
            continue
        filteredbench = copy.deepcopy(be)  # Do NOT modify the old name!
        filteredbench["name"] = regex.sub(replacement, filteredbench["name"])
        filtered["benchmarks"].append(filteredbench)
    return filtered


def get_unique_benchmark_names(json):
    """
    While *keeping* the order, give all the unique 'names' used for benchmarks.
    """
    seen = set()
    uniqued = [
        x["name"]
        for x in json["benchmarks"]
        if x["name"] not in seen and (seen.add(x["name"]) or True)
    ]
    return uniqued


def intersect(list1, list2):
    """
    Given two lists, get a new list consisting of the elements only contained
    in *both of the input lists*, while preserving the ordering.
    """
    return [x for x in list1 if x in list2]


def is_potentially_comparable_benchmark(x):
    return "time_unit" in x and "real_time" in x and "cpu_time" in x


def partition_benchmarks(json1, json2):
    """
    While preserving the ordering, find benchmarks with the same names in
    both of the inputs, and group them.
    (i.e. partition/filter into groups with common name)
    """
    json1_unique_names = get_unique_benchmark_names(json1)
    json2_unique_names = get_unique_benchmark_names(json2)
    names = intersect(json1_unique_names, json2_unique_names)
    partitions = []
    for name in names:
        time_unit = None
        # Pick the time unit from the first entry of the lhs benchmark.
        # We should be careful not to crash with unexpected input.
        for x in json1["benchmarks"]:
            if x["name"] == name and is_potentially_comparable_benchmark(x):
                time_unit = x["time_unit"]
                break
        if time_unit is None:
            continue
        # Filter by name and time unit.
        # All the repetitions are assumed to be comparable.
        lhs = [
            x
            for x in json1["benchmarks"]
            if x["name"] == name and x["time_unit"] == time_unit
        ]
        rhs = [
            x
            for x in json2["benchmarks"]
            if x["name"] == name and x["time_unit"] == time_unit
        ]
        partitions.append([lhs, rhs])
    return partitions


def get_timedelta_field_as_seconds(benchmark, field_name):
    """
    Get value of field_name field of benchmark, which is time with time unit
    time_unit, as time in seconds.
    """
    timedelta = benchmark[field_name]
    time_unit = benchmark.get("time_unit", "s")
    return timedelta * _TIME_UNIT_TO_SECONDS_MULTIPLIER.get(time_unit)


def calculate_geomean(json):
    """
    Extract all real/cpu times from all the benchmarks as seconds,
    and calculate their geomean.
    """
    times = []
    for benchmark in json["benchmarks"]:
        if "run_type" in benchmark and benchmark["run_type"] == "aggregate":
            continue
        times.append(
            [
                get_timedelta_field_as_seconds(benchmark, "real_time"),
                get_timedelta_field_as_seconds(benchmark, "cpu_time"),
            ]
        )
    return gmean(times) if times else array([])


def extract_field(partition, field_name):
    # The count of elements may be different. We want *all* of them.
    lhs = [x[field_name] for x in partition[0]]
    rhs = [x[field_name] for x in partition[1]]
    return [lhs, rhs]


def calc_utest(timings_cpu, timings_time):
    min_rep_cnt = min(
        len(timings_time[0]),
        len(timings_time[1]),
        len(timings_cpu[0]),
        len(timings_cpu[1]),
    )

    # Does *everything* has at least UTEST_MIN_REPETITIONS repetitions?
    if min_rep_cnt < UTEST_MIN_REPETITIONS:
        return False, None, None

    time_pvalue = mannwhitneyu(
        timings_time[0], timings_time[1], alternative="two-sided"
    ).pvalue
    cpu_pvalue = mannwhitneyu(
        timings_cpu[0], timings_cpu[1], alternative="two-sided"
    ).pvalue

    return (min_rep_cnt >= UTEST_OPTIMAL_REPETITIONS), cpu_pvalue, time_pvalue


def print_utest(bc_name, utest, utest_alpha, first_col_width, use_color=True):
    def get_utest_color(pval):
        return BC_FAIL if pval >= utest_alpha else BC_OKGREEN

    # Check if we failed miserably with minimum required repetitions for utest
    if (
        not utest["have_optimal_repetitions"]
        and utest["cpu_pvalue"] is None
        and utest["time_pvalue"] is None
    ):
        return []

    dsc = "U Test, Repetitions: {} vs {}".format(
        utest["nr_of_repetitions"], utest["nr_of_repetitions_other"]
    )
    dsc_color = BC_OKGREEN

    # We still got some results to show but issue a warning about it.
    if not utest["have_optimal_repetitions"]:
        dsc_color = BC_WARNING
        dsc += ". WARNING: Results unreliable! {}+ repetitions recommended.".format(
            UTEST_OPTIMAL_REPETITIONS
        )

    special_str = "{}{:<{}s}{endc}{}{:16.4f}{endc}{}{:16.4f}{endc}{}      {}"

    return [
        color_format(
            use_color,
            special_str,
            BC_HEADER,
            "{}{}".format(bc_name, UTEST_COL_NAME),
            first_col_width,
            get_utest_color(utest["time_pvalue"]),
            utest["time_pvalue"],
            get_utest_color(utest["cpu_pvalue"]),
            utest["cpu_pvalue"],
            dsc_color,
            dsc,
            endc=BC_ENDC,
        )
    ]


def get_difference_report(json1, json2, utest=False):
    """
    Calculate and report the difference between each test of two benchmarks
    runs specified as 'json1' and 'json2'. Output is another json containing
    relevant details for each test run.
    """
    assert utest is True or utest is False

    diff_report = []
    partitions = partition_benchmarks(json1, json2)
    for partition in partitions:
        benchmark_name = partition[0][0]["name"]
        label = partition[0][0]["label"] if "label" in partition[0][0] else ""
        time_unit = partition[0][0]["time_unit"]
        measurements = []
        utest_results = {}
        # Careful, we may have different repetition count.
        for i in range(min(len(partition[0]), len(partition[1]))):
            bn = partition[0][i]
            other_bench = partition[1][i]
            measurements.append(
                {
                    "real_time": bn["real_time"],
                    "cpu_time": bn["cpu_time"],
                    "real_time_other": other_bench["real_time"],
                    "cpu_time_other": other_bench["cpu_time"],
                    "time": calculate_change(
                        bn["real_time"], other_bench["real_time"]
                    ),
                    "cpu": calculate_change(
                        bn["cpu_time"], other_bench["cpu_time"]
                    ),
                }
            )

        # After processing the whole partition, if requested, do the U test.
        if utest:
            timings_cpu = extract_field(partition, "cpu_time")
            timings_time = extract_field(partition, "real_time")
            have_optimal_repetitions, cpu_pvalue, time_pvalue = calc_utest(
                timings_cpu, timings_time
            )
            if cpu_pvalue is not None and time_pvalue is not None:
                utest_results = {
                    "have_optimal_repetitions": have_optimal_repetitions,
                    "cpu_pvalue": cpu_pvalue,
                    "time_pvalue": time_pvalue,
                    "nr_of_repetitions": len(timings_cpu[0]),
                    "nr_of_repetitions_other": len(timings_cpu[1]),
                }

        # Store only if we had any measurements for given benchmark.
        # E.g. partition_benchmarks will filter out the benchmarks having
        # time units which are not compatible with other time units in the
        # benchmark suite.
        if measurements:
            run_type = (
                partition[0][0]["run_type"]
                if "run_type" in partition[0][0]
                else ""
            )
            aggregate_name = (
                partition[0][0]["aggregate_name"]
                if run_type == "aggregate"
                and "aggregate_name" in partition[0][0]
                else ""
            )
            diff_report.append(
                {
                    "name": benchmark_name,
                    "label": label,
                    "measurements": measurements,
                    "time_unit": time_unit,
                    "run_type": run_type,
                    "aggregate_name": aggregate_name,
                    "utest": utest_results,
                }
            )

    lhs_gmean = calculate_geomean(json1)
    rhs_gmean = calculate_geomean(json2)
    if lhs_gmean.any() and rhs_gmean.any():
        diff_report.append(
            {
                "name": "OVERALL_GEOMEAN",
                "label": "",
                "measurements": [
                    {
                        "real_time": lhs_gmean[0],
                        "cpu_time": lhs_gmean[1],
                        "real_time_other": rhs_gmean[0],
                        "cpu_time_other": rhs_gmean[1],
                        "time": calculate_change(lhs_gmean[0], rhs_gmean[0]),
                        "cpu": calculate_change(lhs_gmean[1], rhs_gmean[1]),
                    }
                ],
                "time_unit": "s",
                "run_type": "aggregate",
                "aggregate_name": "geomean",
                "utest": {},
            }
        )

    return diff_report


def print_difference_report(
    json_diff_report,
    include_aggregates_only=False,
    utest=False,
    utest_alpha=0.05,
    use_color=True,
):
    """
    Calculate and report the difference between each test of two benchmarks
    runs specified as 'json1' and 'json2'.
    """
    assert utest is True or utest is False

    def get_color(res):
        if res > 0.05:
            return BC_FAIL
        elif res > -0.07:
            return BC_WHITE
        else:
            return BC_CYAN

    first_col_width = find_longest_name(json_diff_report)
    first_col_width = max(first_col_width, len("Benchmark"))
    first_col_width += len(UTEST_COL_NAME)
    first_line = "{:<{}s}Time             CPU      Time Old      Time New       CPU Old       CPU New".format(
        "Benchmark", 12 + first_col_width
    )
    output_strs = [first_line, "-" * len(first_line)]

    fmt_str = "{}{:<{}s}{endc}{}{:+16.4f}{endc}{}{:+16.4f}{endc}{:14.0f}{:14.0f}{endc}{:14.0f}{:14.0f}"
    for benchmark in json_diff_report:
        # *If* we were asked to only include aggregates,
        # and if it is non-aggregate, then don't print it.
        if (
            not include_aggregates_only
            or "run_type" not in benchmark
            or benchmark["run_type"] == "aggregate"
        ):
            for measurement in benchmark["measurements"]:
                output_strs += [
                    color_format(
                        use_color,
                        fmt_str,
                        BC_HEADER,
                        benchmark["name"],
                        first_col_width,
                        get_color(measurement["time"]),
                        measurement["time"],
                        get_color(measurement["cpu"]),
                        measurement["cpu"],
                        measurement["real_time"],
                        measurement["real_time_other"],
                        measurement["cpu_time"],
                        measurement["cpu_time_other"],
                        endc=BC_ENDC,
                    )
                ]

        # After processing the measurements, if requested and
        # if applicable (e.g. u-test exists for given benchmark),
        # print the U test.
        if utest and benchmark["utest"]:
            output_strs += print_utest(
                benchmark["name"],
                benchmark["utest"],
                utest_alpha=utest_alpha,
                first_col_width=first_col_width,
                use_color=use_color,
            )

    return output_strs


###############################################################################
# Unit tests


class TestGetUniqueBenchmarkNames(unittest.TestCase):
    def load_results(self):
        import json

        testInputs = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "Inputs"
        )
        testOutput = os.path.join(testInputs, "test3_run0.json")
        with open(testOutput, "r") as f:
            json = json.load(f)
        return json

    def test_basic(self):
        expect_lines = [
            "BM_One",
            "BM_Two",
            "short",  # These two are not sorted
            "medium",  # These two are not sorted
        ]
        json = self.load_results()
        output_lines = get_unique_benchmark_names(json)
        print("\n")
        print("\n".join(output_lines))
        self.assertEqual(len(output_lines), len(expect_lines))
        for i in range(0, len(output_lines)):
            self.assertEqual(expect_lines[i], output_lines[i])


class TestReportDifference(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        def load_results():
            import json

            testInputs = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "Inputs"
            )
            testOutput1 = os.path.join(testInputs, "test1_run1.json")
            testOutput2 = os.path.join(testInputs, "test1_run2.json")
            with open(testOutput1, "r") as f:
                json1 = json.load(f)
            with open(testOutput2, "r") as f:
                json2 = json.load(f)
            return json1, json2

        json1, json2 = load_results()
        cls.json_diff_report = get_difference_report(json1, json2)

    def test_json_diff_report_pretty_printing(self):
        expect_lines = [
            ["BM_SameTimes", "+0.0000", "+0.0000", "10", "10", "10", "10"],
            ["BM_2xFaster", "-0.5000", "-0.5000", "50", "25", "50", "25"],
            ["BM_2xSlower", "+1.0000", "+1.0000", "50", "100", "50", "100"],
            [
                "BM_1PercentFaster",
                "-0.0100",
                "-0.0100",
                "100",
                "99",
                "100",
                "99",
            ],
            [
                "BM_1PercentSlower",
                "+0.0100",
                "+0.0100",
                "100",
                "101",
                "100",
                "101",
            ],
            [
                "BM_10PercentFaster",
                "-0.1000",
                "-0.1000",
                "100",
                "90",
                "100",
                "90",
            ],
            [
                "BM_10PercentSlower",
                "+0.1000",
                "+0.1000",
                "100",
                "110",
                "100",
                "110",
            ],
            [
                "BM_100xSlower",
                "+99.0000",
                "+99.0000",
                "100",
                "10000",
                "100",
                "10000",
            ],
            [
                "BM_100xFaster",
                "-0.9900",
                "-0.9900",
                "10000",
                "100",
                "10000",
                "100",
            ],
            [
                "BM_10PercentCPUToTime",
                "+0.1000",
                "-0.1000",
                "100",
                "110",
                "100",
                "90",
            ],
            ["BM_ThirdFaster", "-0.3333", "-0.3334", "100", "67", "100", "67"],
            ["BM_NotBadTimeUnit", "-0.9000", "+0.2000", "0", "0", "0", "1"],
            ["BM_hasLabel", "+0.0000", "+0.0000", "1", "1", "1", "1"],
            ["OVERALL_GEOMEAN", "-0.8113", "-0.7779", "0", "0", "0", "0"],
        ]
        output_lines_with_header = print_difference_report(
            self.json_diff_report, use_color=False
        )
        output_lines = output_lines_with_header[2:]
        print("\n")
        print("\n".join(output_lines_with_header))
        self.assertEqual(len(output_lines), len(expect_lines))
        for i in range(0, len(output_lines)):
            parts = [x for x in output_lines[i].split(" ") if x]
            self.assertEqual(len(parts), 7)
            self.assertEqual(expect_lines[i], parts)

    def test_json_diff_report_output(self):
        expected_output = [
            {
                "name": "BM_SameTimes",
                "label": "",
                "measurements": [
                    {
                        "time": 0.0000,
                        "cpu": 0.0000,
                        "real_time": 10,
                        "real_time_other": 10,
                        "cpu_time": 10,
                        "cpu_time_other": 10,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_2xFaster",
                "label": "",
                "measurements": [
                    {
                        "time": -0.5000,
                        "cpu": -0.5000,
                        "real_time": 50,
                        "real_time_other": 25,
                        "cpu_time": 50,
                        "cpu_time_other": 25,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_2xSlower",
                "label": "",
                "measurements": [
                    {
                        "time": 1.0000,
                        "cpu": 1.0000,
                        "real_time": 50,
                        "real_time_other": 100,
                        "cpu_time": 50,
                        "cpu_time_other": 100,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_1PercentFaster",
                "label": "",
                "measurements": [
                    {
                        "time": -0.0100,
                        "cpu": -0.0100,
                        "real_time": 100,
                        "real_time_other": 98.9999999,
                        "cpu_time": 100,
                        "cpu_time_other": 98.9999999,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_1PercentSlower",
                "label": "",
                "measurements": [
                    {
                        "time": 0.0100,
                        "cpu": 0.0100,
                        "real_time": 100,
                        "real_time_other": 101,
                        "cpu_time": 100,
                        "cpu_time_other": 101,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_10PercentFaster",
                "label": "",
                "measurements": [
                    {
                        "time": -0.1000,
                        "cpu": -0.1000,
                        "real_time": 100,
                        "real_time_other": 90,
                        "cpu_time": 100,
                        "cpu_time_other": 90,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_10PercentSlower",
                "label": "",
                "measurements": [
                    {
                        "time": 0.1000,
                        "cpu": 0.1000,
                        "real_time": 100,
                        "real_time_other": 110,
                        "cpu_time": 100,
                        "cpu_time_other": 110,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_100xSlower",
                "label": "",
                "measurements": [
                    {
                        "time": 99.0000,
                        "cpu": 99.0000,
                        "real_time": 100,
                        "real_time_other": 10000,
                        "cpu_time": 100,
                        "cpu_time_other": 10000,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_100xFaster",
                "label": "",
                "measurements": [
                    {
                        "time": -0.9900,
                        "cpu": -0.9900,
                        "real_time": 10000,
                        "real_time_other": 100,
                        "cpu_time": 10000,
                        "cpu_time_other": 100,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_10PercentCPUToTime",
                "label": "",
                "measurements": [
                    {
                        "time": 0.1000,
                        "cpu": -0.1000,
                        "real_time": 100,
                        "real_time_other": 110,
                        "cpu_time": 100,
                        "cpu_time_other": 90,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_ThirdFaster",
                "label": "",
                "measurements": [
                    {
                        "time": -0.3333,
                        "cpu": -0.3334,
                        "real_time": 100,
                        "real_time_other": 67,
                        "cpu_time": 100,
                        "cpu_time_other": 67,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_NotBadTimeUnit",
                "label": "",
                "measurements": [
                    {
                        "time": -0.9000,
                        "cpu": 0.2000,
                        "real_time": 0.4,
                        "real_time_other": 0.04,
                        "cpu_time": 0.5,
                        "cpu_time_other": 0.6,
                    }
                ],
                "time_unit": "s",
                "utest": {},
            },
            {
                "name": "BM_hasLabel",
                "label": "a label",
                "measurements": [
                    {
                        "time": 0.0000,
                        "cpu": 0.0000,
                        "real_time": 1,
                        "real_time_other": 1,
                        "cpu_time": 1,
                        "cpu_time_other": 1,
                    }
                ],
                "time_unit": "s",
                "utest": {},
            },
            {
                "name": "OVERALL_GEOMEAN",
                "label": "",
                "measurements": [
                    {
                        "real_time": 3.1622776601683826e-06,
                        "cpu_time": 3.2130844755623912e-06,
                        "real_time_other": 1.9768988699420897e-07,
                        "cpu_time_other": 2.397447755209533e-07,
                        "time": -0.8112976497120911,
                        "cpu": -0.7778551721181174,
                    }
                ],
                "time_unit": "s",
                "run_type": "aggregate",
                "aggregate_name": "geomean",
                "utest": {},
            },
        ]
        self.assertEqual(len(self.json_diff_report), len(expected_output))
        for out, expected in zip(self.json_diff_report, expected_output):
            self.assertEqual(out["name"], expected["name"])
            self.assertEqual(out["label"], expected["label"])
            self.assertEqual(out["time_unit"], expected["time_unit"])
            assert_utest(self, out, expected)
            assert_measurements(self, out, expected)


class TestReportDifferenceBetweenFamilies(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        def load_result():
            import json

            testInputs = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "Inputs"
            )
            testOutput = os.path.join(testInputs, "test2_run.json")
            with open(testOutput, "r") as f:
                json = json.load(f)
            return json

        json = load_result()
        json1 = filter_benchmark(json, "BM_Z.ro", ".")
        json2 = filter_benchmark(json, "BM_O.e", ".")
        cls.json_diff_report = get_difference_report(json1, json2)

    def test_json_diff_report_pretty_printing(self):
        expect_lines = [
            [".", "-0.5000", "-0.5000", "10", "5", "10", "5"],
            ["./4", "-0.5000", "-0.5000", "40", "20", "40", "20"],
            ["Prefix/.", "-0.5000", "-0.5000", "20", "10", "20", "10"],
            ["Prefix/./3", "-0.5000", "-0.5000", "30", "15", "30", "15"],
            ["OVERALL_GEOMEAN", "-0.5000", "-0.5000", "0", "0", "0", "0"],
        ]
        output_lines_with_header = print_difference_report(
            self.json_diff_report, use_color=False
        )
        output_lines = output_lines_with_header[2:]
        print("\n")
        print("\n".join(output_lines_with_header))
        self.assertEqual(len(output_lines), len(expect_lines))
        for i in range(0, len(output_lines)):
            parts = [x for x in output_lines[i].split(" ") if x]
            self.assertEqual(len(parts), 7)
            self.assertEqual(expect_lines[i], parts)

    def test_json_diff_report(self):
        expected_output = [
            {
                "name": ".",
                "measurements": [
                    {
                        "time": -0.5,
                        "cpu": -0.5,
                        "real_time": 10,
                        "real_time_other": 5,
                        "cpu_time": 10,
                        "cpu_time_other": 5,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "./4",
                "measurements": [
                    {
                        "time": -0.5,
                        "cpu": -0.5,
                        "real_time": 40,
                        "real_time_other": 20,
                        "cpu_time": 40,
                        "cpu_time_other": 20,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "Prefix/.",
                "measurements": [
                    {
                        "time": -0.5,
                        "cpu": -0.5,
                        "real_time": 20,
                        "real_time_other": 10,
                        "cpu_time": 20,
                        "cpu_time_other": 10,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "Prefix/./3",
                "measurements": [
                    {
                        "time": -0.5,
                        "cpu": -0.5,
                        "real_time": 30,
                        "real_time_other": 15,
                        "cpu_time": 30,
                        "cpu_time_other": 15,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "OVERALL_GEOMEAN",
                "measurements": [
                    {
                        "real_time": 2.213363839400641e-08,
                        "cpu_time": 2.213363839400641e-08,
                        "real_time_other": 1.1066819197003185e-08,
                        "cpu_time_other": 1.1066819197003185e-08,
                        "time": -0.5000000000000009,
                        "cpu": -0.5000000000000009,
                    }
                ],
                "time_unit": "s",
                "run_type": "aggregate",
                "aggregate_name": "geomean",
                "utest": {},
            },
        ]
        self.assertEqual(len(self.json_diff_report), len(expected_output))
        for out, expected in zip(self.json_diff_report, expected_output):
            self.assertEqual(out["name"], expected["name"])
            self.assertEqual(out["time_unit"], expected["time_unit"])
            assert_utest(self, out, expected)
            assert_measurements(self, out, expected)


class TestReportDifferenceWithUTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        def load_results():
            import json

            testInputs = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "Inputs"
            )
            testOutput1 = os.path.join(testInputs, "test3_run0.json")
            testOutput2 = os.path.join(testInputs, "test3_run1.json")
            with open(testOutput1, "r") as f:
                json1 = json.load(f)
            with open(testOutput2, "r") as f:
                json2 = json.load(f)
            return json1, json2

        json1, json2 = load_results()
        cls.json_diff_report = get_difference_report(json1, json2, utest=True)

    def test_json_diff_report_pretty_printing(self):
        expect_lines = [
            ["BM_One", "-0.1000", "+0.1000", "10", "9", "100", "110"],
            ["BM_Two", "+0.1111", "-0.0111", "9", "10", "90", "89"],
            ["BM_Two", "-0.1250", "-0.1628", "8", "7", "86", "72"],
            [
                "BM_Two_pvalue",
                "1.0000",
                "0.6667",
                "U",
                "Test,",
                "Repetitions:",
                "2",
                "vs",
                "2.",
                "WARNING:",
                "Results",
                "unreliable!",
                "9+",
                "repetitions",
                "recommended.",
            ],
            ["short", "-0.1250", "-0.0625", "8", "7", "80", "75"],
            ["short", "-0.4325", "-0.1351", "8", "5", "77", "67"],
            [
                "short_pvalue",
                "0.7671",
                "0.2000",
                "U",
                "Test,",
                "Repetitions:",
                "2",
                "vs",
                "3.",
                "WARNING:",
                "Results",
                "unreliable!",
                "9+",
                "repetitions",
                "recommended.",
            ],
            ["medium", "-0.3750", "-0.3375", "8", "5", "80", "53"],
            ["OVERALL_GEOMEAN", "+1.6405", "-0.6985", "0", "0", "0", "0"],
        ]
        output_lines_with_header = print_difference_report(
            self.json_diff_report, utest=True, utest_alpha=0.05, use_color=False
        )
        output_lines = output_lines_with_header[2:]
        print("\n")
        print("\n".join(output_lines_with_header))
        self.assertEqual(len(output_lines), len(expect_lines))
        for i in range(0, len(output_lines)):
            parts = [x for x in output_lines[i].split(" ") if x]
            self.assertEqual(expect_lines[i], parts)

    def test_json_diff_report_pretty_printing_aggregates_only(self):
        expect_lines = [
            ["BM_One", "-0.1000", "+0.1000", "10", "9", "100", "110"],
            [
                "BM_Two_pvalue",
                "1.0000",
                "0.6667",
                "U",
                "Test,",
                "Repetitions:",
                "2",
                "vs",
                "2.",
                "WARNING:",
                "Results",
                "unreliable!",
                "9+",
                "repetitions",
                "recommended.",
            ],
            ["short", "-0.1250", "-0.0625", "8", "7", "80", "75"],
            ["short", "-0.4325", "-0.1351", "8", "5", "77", "67"],
            [
                "short_pvalue",
                "0.7671",
                "0.2000",
                "U",
                "Test,",
                "Repetitions:",
                "2",
                "vs",
                "3.",
                "WARNING:",
                "Results",
                "unreliable!",
                "9+",
                "repetitions",
                "recommended.",
            ],
            ["OVERALL_GEOMEAN", "+1.6405", "-0.6985", "0", "0", "0", "0"],
        ]
        output_lines_with_header = print_difference_report(
            self.json_diff_report,
            include_aggregates_only=True,
            utest=True,
            utest_alpha=0.05,
            use_color=False,
        )
        output_lines = output_lines_with_header[2:]
        print("\n")
        print("\n".join(output_lines_with_header))
        self.assertEqual(len(output_lines), len(expect_lines))
        for i in range(0, len(output_lines)):
            parts = [x for x in output_lines[i].split(" ") if x]
            self.assertEqual(expect_lines[i], parts)

    def test_json_diff_report(self):
        expected_output = [
            {
                "name": "BM_One",
                "measurements": [
                    {
                        "time": -0.1,
                        "cpu": 0.1,
                        "real_time": 10,
                        "real_time_other": 9,
                        "cpu_time": 100,
                        "cpu_time_other": 110,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_Two",
                "measurements": [
                    {
                        "time": 0.1111111111111111,
                        "cpu": -0.011111111111111112,
                        "real_time": 9,
                        "real_time_other": 10,
                        "cpu_time": 90,
                        "cpu_time_other": 89,
                    },
                    {
                        "time": -0.125,
                        "cpu": -0.16279069767441862,
                        "real_time": 8,
                        "real_time_other": 7,
                        "cpu_time": 86,
                        "cpu_time_other": 72,
                    },
                ],
                "time_unit": "ns",
                "utest": {
                    "have_optimal_repetitions": False,
                    "cpu_pvalue": 0.6666666666666666,
                    "time_pvalue": 1.0,
                },
            },
            {
                "name": "short",
                "measurements": [
                    {
                        "time": -0.125,
                        "cpu": -0.0625,
                        "real_time": 8,
                        "real_time_other": 7,
                        "cpu_time": 80,
                        "cpu_time_other": 75,
                    },
                    {
                        "time": -0.4325,
                        "cpu": -0.13506493506493514,
                        "real_time": 8,
                        "real_time_other": 4.54,
                        "cpu_time": 77,
                        "cpu_time_other": 66.6,
                    },
                ],
                "time_unit": "ns",
                "utest": {
                    "have_optimal_repetitions": False,
                    "cpu_pvalue": 0.2,
                    "time_pvalue": 0.7670968684102772,
                },
            },
            {
                "name": "medium",
                "measurements": [
                    {
                        "time": -0.375,
                        "cpu": -0.3375,
                        "real_time": 8,
                        "real_time_other": 5,
                        "cpu_time": 80,
                        "cpu_time_other": 53,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "OVERALL_GEOMEAN",
                "measurements": [
                    {
                        "real_time": 8.48528137423858e-09,
                        "cpu_time": 8.441336246629233e-08,
                        "real_time_other": 2.2405267593145244e-08,
                        "cpu_time_other": 2.5453661413660466e-08,
                        "time": 1.6404861082353634,
                        "cpu": -0.6984640740519662,
                    }
                ],
                "time_unit": "s",
                "run_type": "aggregate",
                "aggregate_name": "geomean",
                "utest": {},
            },
        ]
        self.assertEqual(len(self.json_diff_report), len(expected_output))
        for out, expected in zip(self.json_diff_report, expected_output):
            self.assertEqual(out["name"], expected["name"])
            self.assertEqual(out["time_unit"], expected["time_unit"])
            assert_utest(self, out, expected)
            assert_measurements(self, out, expected)


class TestReportDifferenceWithUTestWhileDisplayingAggregatesOnly(
    unittest.TestCase
):
    @classmethod
    def setUpClass(cls):
        def load_results():
            import json

            testInputs = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "Inputs"
            )
            testOutput1 = os.path.join(testInputs, "test3_run0.json")
            testOutput2 = os.path.join(testInputs, "test3_run1.json")
            with open(testOutput1, "r") as f:
                json1 = json.load(f)
            with open(testOutput2, "r") as f:
                json2 = json.load(f)
            return json1, json2

        json1, json2 = load_results()
        cls.json_diff_report = get_difference_report(json1, json2, utest=True)

    def test_json_diff_report_pretty_printing(self):
        expect_lines = [
            ["BM_One", "-0.1000", "+0.1000", "10", "9", "100", "110"],
            ["BM_Two", "+0.1111", "-0.0111", "9", "10", "90", "89"],
            ["BM_Two", "-0.1250", "-0.1628", "8", "7", "86", "72"],
            [
                "BM_Two_pvalue",
                "1.0000",
                "0.6667",
                "U",
                "Test,",
                "Repetitions:",
                "2",
                "vs",
                "2.",
                "WARNING:",
                "Results",
                "unreliable!",
                "9+",
                "repetitions",
                "recommended.",
            ],
            ["short", "-0.1250", "-0.0625", "8", "7", "80", "75"],
            ["short", "-0.4325", "-0.1351", "8", "5", "77", "67"],
            [
                "short_pvalue",
                "0.7671",
                "0.2000",
                "U",
                "Test,",
                "Repetitions:",
                "2",
                "vs",
                "3.",
                "WARNING:",
                "Results",
                "unreliable!",
                "9+",
                "repetitions",
                "recommended.",
            ],
            ["medium", "-0.3750", "-0.3375", "8", "5", "80", "53"],
            ["OVERALL_GEOMEAN", "+1.6405", "-0.6985", "0", "0", "0", "0"],
        ]
        output_lines_with_header = print_difference_report(
            self.json_diff_report, utest=True, utest_alpha=0.05, use_color=False
        )
        output_lines = output_lines_with_header[2:]
        print("\n")
        print("\n".join(output_lines_with_header))
        self.assertEqual(len(output_lines), len(expect_lines))
        for i in range(0, len(output_lines)):
            parts = [x for x in output_lines[i].split(" ") if x]
            self.assertEqual(expect_lines[i], parts)

    def test_json_diff_report(self):
        expected_output = [
            {
                "name": "BM_One",
                "measurements": [
                    {
                        "time": -0.1,
                        "cpu": 0.1,
                        "real_time": 10,
                        "real_time_other": 9,
                        "cpu_time": 100,
                        "cpu_time_other": 110,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            },
            {
                "name": "BM_Two",
                "measurements": [
                    {
                        "time": 0.1111111111111111,
                        "cpu": -0.011111111111111112,
                        "real_time": 9,
                        "real_time_other": 10,
                        "cpu_time": 90,
                        "cpu_time_other": 89,
                    },
                    {
                        "time": -0.125,
                        "cpu": -0.16279069767441862,
                        "real_time": 8,
                        "real_time_other": 7,
                        "cpu_time": 86,
                        "cpu_time_other": 72,
                    },
                ],
                "time_unit": "ns",
                "utest": {
                    "have_optimal_repetitions": False,
                    "cpu_pvalue": 0.6666666666666666,
                    "time_pvalue": 1.0,
                },
            },
            {
                "name": "short",
                "measurements": [
                    {
                        "time": -0.125,
                        "cpu": -0.0625,
                        "real_time": 8,
                        "real_time_other": 7,
                        "cpu_time": 80,
                        "cpu_time_other": 75,
                    },
                    {
                        "time": -0.4325,
                        "cpu": -0.13506493506493514,
                        "real_time": 8,
                        "real_time_other": 4.54,
                        "cpu_time": 77,
                        "cpu_time_other": 66.6,
                    },
                ],
                "time_unit": "ns",
                "utest": {
                    "have_optimal_repetitions": False,
                    "cpu_pvalue": 0.2,
                    "time_pvalue": 0.7670968684102772,
                },
            },
            {
                "name": "medium",
                "measurements": [
                    {
                        "real_time_other": 5,
                        "cpu_time": 80,
                        "time": -0.375,
                        "real_time": 8,
                        "cpu_time_other": 53,
                        "cpu": -0.3375,
                    }
                ],
                "utest": {},
                "time_unit": "ns",
                "aggregate_name": "",
            },
            {
                "name": "OVERALL_GEOMEAN",
                "measurements": [
                    {
                        "real_time": 8.48528137423858e-09,
                        "cpu_time": 8.441336246629233e-08,
                        "real_time_other": 2.2405267593145244e-08,
                        "cpu_time_other": 2.5453661413660466e-08,
                        "time": 1.6404861082353634,
                        "cpu": -0.6984640740519662,
                    }
                ],
                "time_unit": "s",
                "run_type": "aggregate",
                "aggregate_name": "geomean",
                "utest": {},
            },
        ]
        self.assertEqual(len(self.json_diff_report), len(expected_output))
        for out, expected in zip(self.json_diff_report, expected_output):
            self.assertEqual(out["name"], expected["name"])
            self.assertEqual(out["time_unit"], expected["time_unit"])
            assert_utest(self, out, expected)
            assert_measurements(self, out, expected)


class TestReportDifferenceForPercentageAggregates(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        def load_results():
            import json

            testInputs = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "Inputs"
            )
            testOutput1 = os.path.join(testInputs, "test4_run0.json")
            testOutput2 = os.path.join(testInputs, "test4_run1.json")
            with open(testOutput1, "r") as f:
                json1 = json.load(f)
            with open(testOutput2, "r") as f:
                json2 = json.load(f)
            return json1, json2

        json1, json2 = load_results()
        cls.json_diff_report = get_difference_report(json1, json2, utest=True)

    def test_json_diff_report_pretty_printing(self):
        expect_lines = [["whocares", "-0.5000", "+0.5000", "0", "0", "0", "0"]]
        output_lines_with_header = print_difference_report(
            self.json_diff_report, utest=True, utest_alpha=0.05, use_color=False
        )
        output_lines = output_lines_with_header[2:]
        print("\n")
        print("\n".join(output_lines_with_header))
        self.assertEqual(len(output_lines), len(expect_lines))
        for i in range(0, len(output_lines)):
            parts = [x for x in output_lines[i].split(" ") if x]
            self.assertEqual(expect_lines[i], parts)

    def test_json_diff_report(self):
        expected_output = [
            {
                "name": "whocares",
                "measurements": [
                    {
                        "time": -0.5,
                        "cpu": 0.5,
                        "real_time": 0.01,
                        "real_time_other": 0.005,
                        "cpu_time": 0.10,
                        "cpu_time_other": 0.15,
                    }
                ],
                "time_unit": "ns",
                "utest": {},
            }
        ]
        self.assertEqual(len(self.json_diff_report), len(expected_output))
        for out, expected in zip(self.json_diff_report, expected_output):
            self.assertEqual(out["name"], expected["name"])
            self.assertEqual(out["time_unit"], expected["time_unit"])
            assert_utest(self, out, expected)
            assert_measurements(self, out, expected)


class TestReportSorting(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        def load_result():
            import json

            testInputs = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "Inputs"
            )
            testOutput = os.path.join(testInputs, "test4_run.json")
            with open(testOutput, "r") as f:
                json = json.load(f)
            return json

        cls.json = load_result()

    def test_json_diff_report_pretty_printing(self):
        import util

        expected_names = [
            "99 family 0 instance 0 repetition 0",
            "98 family 0 instance 0 repetition 1",
            "97 family 0 instance 0 aggregate",
            "96 family 0 instance 1 repetition 0",
            "95 family 0 instance 1 repetition 1",
            "94 family 0 instance 1 aggregate",
            "93 family 1 instance 0 repetition 0",
            "92 family 1 instance 0 repetition 1",
            "91 family 1 instance 0 aggregate",
            "90 family 1 instance 1 repetition 0",
            "89 family 1 instance 1 repetition 1",
            "88 family 1 instance 1 aggregate",
        ]

        for n in range(len(self.json["benchmarks"]) ** 2):
            random.shuffle(self.json["benchmarks"])
            sorted_benchmarks = util.sort_benchmark_results(self.json)[
                "benchmarks"
            ]
            self.assertEqual(len(expected_names), len(sorted_benchmarks))
            for out, expected in zip(sorted_benchmarks, expected_names):
                self.assertEqual(out["name"], expected)


class TestReportDifferenceWithUTestWhileDisplayingAggregatesOnly2(
    unittest.TestCase
):
    @classmethod
    def setUpClass(cls):
        def load_results():
            import json

            testInputs = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "Inputs"
            )
            testOutput1 = os.path.join(testInputs, "test5_run0.json")
            testOutput2 = os.path.join(testInputs, "test5_run1.json")
            with open(testOutput1, "r") as f:
                json1 = json.load(f)
                json1["benchmarks"] = [
                    json1["benchmarks"][0] for i in range(1000)
                ]
            with open(testOutput2, "r") as f:
                json2 = json.load(f)
                json2["benchmarks"] = [
                    json2["benchmarks"][0] for i in range(1000)
                ]
            return json1, json2

        json1, json2 = load_results()
        cls.json_diff_report = get_difference_report(json1, json2, utest=True)

    def test_json_diff_report_pretty_printing(self):
        expect_line = [
            "BM_ManyRepetitions_pvalue",
            "0.0000",
            "0.0000",
            "U",
            "Test,",
            "Repetitions:",
            "1000",
            "vs",
            "1000",
        ]
        output_lines_with_header = print_difference_report(
            self.json_diff_report, utest=True, utest_alpha=0.05, use_color=False
        )
        output_lines = output_lines_with_header[2:]
        found = False
        for i in range(0, len(output_lines)):
            parts = [x for x in output_lines[i].split(" ") if x]
            found = expect_line == parts
            if found:
                break
        self.assertTrue(found)

    def test_json_diff_report(self):
        expected_output = [
            {
                "name": "BM_ManyRepetitions",
                "label": "",
                "time_unit": "s",
                "run_type": "",
                "aggregate_name": "",
                "utest": {
                    "have_optimal_repetitions": True,
                    "cpu_pvalue": 0.0,
                    "time_pvalue": 0.0,
                    "nr_of_repetitions": 1000,
                    "nr_of_repetitions_other": 1000,
                },
            },
            {
                "name": "OVERALL_GEOMEAN",
                "label": "",
                "measurements": [
                    {
                        "real_time": 1.0,
                        "cpu_time": 1000.000000000069,
                        "real_time_other": 1000.000000000069,
                        "cpu_time_other": 1.0,
                        "time": 999.000000000069,
                        "cpu": -0.9990000000000001,
                    }
                ],
                "time_unit": "s",
                "run_type": "aggregate",
                "aggregate_name": "geomean",
                "utest": {},
            },
        ]
        self.assertEqual(len(self.json_diff_report), len(expected_output))
        for out, expected in zip(self.json_diff_report, expected_output):
            self.assertEqual(out["name"], expected["name"])
            self.assertEqual(out["time_unit"], expected["time_unit"])
            assert_utest(self, out, expected)


def assert_utest(unittest_instance, lhs, rhs):
    if lhs["utest"]:
        unittest_instance.assertAlmostEqual(
            lhs["utest"]["cpu_pvalue"], rhs["utest"]["cpu_pvalue"]
        )
        unittest_instance.assertAlmostEqual(
            lhs["utest"]["time_pvalue"], rhs["utest"]["time_pvalue"]
        )
        unittest_instance.assertEqual(
            lhs["utest"]["have_optimal_repetitions"],
            rhs["utest"]["have_optimal_repetitions"],
        )
    else:
        # lhs is empty. assert if rhs is not.
        unittest_instance.assertEqual(lhs["utest"], rhs["utest"])


def assert_measurements(unittest_instance, lhs, rhs):
    for m1, m2 in zip(lhs["measurements"], rhs["measurements"]):
        unittest_instance.assertEqual(m1["real_time"], m2["real_time"])
        unittest_instance.assertEqual(m1["cpu_time"], m2["cpu_time"])
        # m1['time'] and m1['cpu'] hold values which are being calculated,
        # and therefore we must use almost-equal pattern.
        unittest_instance.assertAlmostEqual(m1["time"], m2["time"], places=4)
        unittest_instance.assertAlmostEqual(m1["cpu"], m2["cpu"], places=4)


if __name__ == "__main__":
    unittest.main()

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# kate: tab-width: 4; replace-tabs on; indent-width 4; tab-indents: off;
# kate: indent-mode python; remove-trailing-spaces modified;



import pytest
import numpy as np

from nwsspc.sharp.calc import layer
from nwsspc.sharp.calc import constants


def test_height_layer_construction():
    lyr1 = layer.HeightLayer(0, 3000)
    assert (lyr1.bottom == 0 and lyr1.top == 3000)

    with pytest.raises(ValueError):
        layer.HeightLayer(3000, 0)

    with pytest.raises(ValueError):
        layer.HeightLayer(0, np.nan)

    with pytest.raises(ValueError):
        layer.HeightLayer(np.nan, 3000)

    with pytest.raises(ValueError):
        layer.HeightLayer(np.nan, np.nan)

    with pytest.raises(ValueError):
        layer.HeightLayer(np.inf, 0)

    with pytest.raises(ValueError):
        layer.HeightLayer(0, np.inf)

    with pytest.raises(ValueError):
        layer.HeightLayer(np.inf, np.inf)

    with pytest.raises(ValueError):
        layer.HeightLayer(np.inf, np.nan)

    with pytest.raises(ValueError):
        layer.HeightLayer(constants.MISSING, 3000.0)

    with pytest.raises(ValueError):
        layer.HeightLayer(0.0, constants.MISSING)

    lyr1 = layer.HeightLayer()
    assert (lyr1.bottom == constants.MISSING)
    assert (lyr1.top == constants.MISSING)


def test_pressure_layer_construction():
    lyr1 = layer.PressureLayer(100000, 10000)
    assert (lyr1.bottom == 100000 and lyr1.top == 10000)

    with pytest.raises(ValueError):
        layer.PressureLayer(10000, 100000)

    with pytest.raises(ValueError):
        layer.PressureLayer(0, np.nan)

    with pytest.raises(ValueError):
        layer.PressureLayer(np.nan, 3000)

    with pytest.raises(ValueError):
        layer.PressureLayer(np.nan, np.nan)

    with pytest.raises(ValueError):
        layer.PressureLayer(np.inf, 0)

    with pytest.raises(ValueError):
        layer.PressureLayer(0, np.inf)

    with pytest.raises(ValueError):
        layer.PressureLayer(np.inf, np.inf)

    with pytest.raises(ValueError):
        layer.PressureLayer(np.inf, np.nan)

    with pytest.raises(ValueError):
        layer.PressureLayer(constants.MISSING, 50000.0)

    with pytest.raises(ValueError):
        layer.PressureLayer(100000.0, constants.MISSING)

    lyr1 = layer.PressureLayer()
    assert (lyr1.bottom == constants.MISSING)
    assert (lyr1.top == constants.MISSING)


def test_layer_conversion():
    pres = np.arange(10000, 110000, 10000)[::-1]
    hght = np.array([0.0, 500.0, 1500.0, 2500.0, 4000.0,
                    5500.0, 7500.0, 8500.0, 10500.0, 12500.0])

    hght_lyr = layer.HeightLayer(0.0, 3000.0)
    pres_lyr = layer.PressureLayer(100000.0, 75000.0)

    new_pres_lyr = layer.height_layer_to_pressure(hght_lyr, pres, hght)
    new_hght_lyr = layer.pressure_layer_to_height(pres_lyr, pres, hght)

    assert (new_pres_lyr.bottom == 100000.0)
    assert (new_hght_lyr.bottom == 0.0)
    assert (new_pres_lyr.top == pytest.approx(66666.7))
    assert (new_hght_lyr.top == pytest.approx(1983.3248))

    # Check out of bounds behavior
    h_oob1 = layer.HeightLayer(-100, 250.0)
    h_oob2 = layer.HeightLayer(11500.0, 14000.0)
    p_oob1 = layer.PressureLayer(115000.0, 90000.0)
    p_oob2 = layer.PressureLayer(15000.0, 5000.0)

    oob1 = layer.height_layer_to_pressure(h_oob1, pres, hght)
    oob2 = layer.height_layer_to_pressure(h_oob2, pres, hght)
    oob3 = layer.pressure_layer_to_height(p_oob1, pres, hght)
    oob4 = layer.pressure_layer_to_height(p_oob2, pres, hght)

    assert (oob1.bottom == constants.MISSING)
    assert (oob2.bottom == constants.MISSING)
    assert (oob3.bottom == constants.MISSING)
    assert (oob4.bottom == constants.MISSING)
    assert (oob1.top == constants.MISSING)
    assert (oob2.top == constants.MISSING)
    assert (oob3.top == constants.MISSING)
    assert (oob4.top == constants.MISSING)

    # Make sure the enforcement of 1D arrays works
    with pytest.raises(TypeError):
        layer.height_layer_to_pressure(hght_lyr, pres.reshape((5, 2)), hght)

    with pytest.raises(TypeError):
        layer.pressure_layer_to_height(pres_lyr, pres.reshape((5, 2)), hght)


def test_layer_bounds_check():
    pres = np.arange(10000, 110000, 10000)[::-1]

    p_oob1 = layer.PressureLayer(100000.0, 5000.0)
    p_oob2 = layer.PressureLayer(110000.0, 10000.0)
    p_oob3 = layer.PressureLayer(110000.0, 5000.0)
    p_oob4 = layer.PressureLayer(5000.0, 2500.0)
    p_oob5 = layer.PressureLayer(115000.0, 110000.0)

    idx1 = layer.get_layer_index(p_oob1, pres)
    idx2 = layer.get_layer_index(p_oob2, pres)
    idx3 = layer.get_layer_index(p_oob3, pres)
    idx4 = layer.get_layer_index(p_oob4, pres)
    idx5 = layer.get_layer_index(p_oob5, pres)

    assert (idx1.kbot == 0)
    assert (idx1.ktop == 9)
    assert (idx2.kbot == 0)
    assert (idx2.ktop == 9)
    assert (idx3.kbot == 0)
    assert (idx3.ktop == 9)
    assert (idx4.kbot == 9)
    assert (idx4.ktop == 9)
    assert (idx5.kbot == 0)
    assert (idx5.ktop == 0)


def test_pressure_layer_min():
    pres = np.arange(10000, 110000, 10000)[::-1]
    data = np.array([0, 0, 0, 0, -10, 0, 0, 0, 0, 0])

    lyr1 = layer.PressureLayer(100000.0, 60000.0)  # min at top of layer
    lyr2 = layer.PressureLayer(110000.0, 50000.0)  # out of bounds recovery
    lyr3 = layer.PressureLayer(100000.0, 85000.0)  # min is not in layer
    lyr4 = layer.PressureLayer(80000.0, 65000.0)  # min is just above layer
    lyr5 = layer.PressureLayer(55000.0, 40000.0)  # min is just below layer

    assert (layer.layer_min(lyr1, pres, data) == (-10.0, 60000.0))
    assert (layer.layer_min(lyr2, pres, data) == (-10.0, 60000.0))
    assert (layer.layer_min(lyr3, pres, data) == (0.0, 100000.0))
    assert (layer.layer_min(lyr4, pres, data) ==
            (pytest.approx(-4.807509899, abs=1e-4), 65000.0))
    assert (layer.layer_min(lyr5, pres, data) ==
            (pytest.approx(-5.22757434, abs=1e-4), 55000.0))


def test_pressure_layer_max():
    pres = np.arange(10000, 110000, 10000)[::-1]
    data = np.array([0, 0, 0, 0, 10, 0, 0, 0, 0, 0])

    lyr1 = layer.PressureLayer(100000.0, 60000.0)  # min at top of layer
    lyr2 = layer.PressureLayer(110000.0, 50000.0)  # out of bounds recovery
    lyr3 = layer.PressureLayer(100000.0, 85000.0)  # min is not in layer
    lyr4 = layer.PressureLayer(80000.0, 65000.0)  # min is just above layer
    lyr5 = layer.PressureLayer(55000.0, 40000.0)  # min is just below layer

    assert (layer.layer_max(lyr1, pres, data) == (10.0, 60000.0))
    assert (layer.layer_max(lyr2, pres, data) == (10.0, 60000.0))
    assert (layer.layer_max(lyr3, pres, data) == (0.0, 100000.0))
    assert (layer.layer_max(lyr4, pres, data) ==
            (pytest.approx(4.807509899, abs=1e-4), 65000.0))
    assert (layer.layer_max(lyr5, pres, data) ==
            (pytest.approx(5.22757434, abs=1e-4), 55000.0))


def test_height_layer_min():
    hght = np.arange(1000.0, 11000.0, 1000.0)
    data = np.array([0, 0, 0, 0, -10, 0, 0, 0, 0, 0])

    lyr1 = layer.HeightLayer(1000.0, 6000.0)  # max at top of layer
    lyr2 = layer.HeightLayer(-100.0, 20000.0)  # out of bounds recovery
    lyr3 = layer.HeightLayer(6000.0, 10000.0)  # max is not in layer
    lyr4 = layer.HeightLayer(1000.0, 4750.0)  # max is just above layer
    lyr5 = layer.HeightLayer(5250.0, 9000.0)  # max is just below layer

    assert (layer.layer_min(lyr1, hght, data) == (-10.0, 5000.0))
    assert (layer.layer_min(lyr2, hght, data) == (-10.0, 5000.0))
    assert (layer.layer_min(lyr3, hght, data) == (0.0, 6000.0))
    assert (layer.layer_min(lyr4, hght, data) == (-7.5, 4750.0))
    assert (layer.layer_min(lyr5, hght, data) == (-7.5, 5250.0))


def test_height_layer_max():
    hght = np.arange(1000.0, 11000.0, 1000.0)
    data = np.array([0, 0, 0, 0, 10, 0, 0, 0, 0, 0])

    lyr1 = layer.HeightLayer(1000.0, 6000.0)  # max at top of layer
    lyr2 = layer.HeightLayer(-100.0, 20000.0)  # out of bounds recovery
    lyr3 = layer.HeightLayer(6000.0, 10000.0)  # max is not in layer
    lyr4 = layer.HeightLayer(1000.0, 4750.0)  # max is just above layer
    lyr5 = layer.HeightLayer(5250.0, 9000.0)  # max is just below layer

    assert (layer.layer_max(lyr1, hght, data) == (10.0, 5000.0))
    assert (layer.layer_max(lyr2, hght, data) == (10.0, 5000.0))
    assert (layer.layer_max(lyr3, hght, data) == (0.0, 6000.0))
    assert (layer.layer_max(lyr4, hght, data) == (7.5, 4750.0))
    assert (layer.layer_max(lyr5, hght, data) == (7.5, 5250.0))


def test_pressure_layer_mean():
    pres = np.arange(10000, 110000, 10000)[::-1]
    data = np.array([0, 0, 0, 0, 10.0, 0, 0, 0, 0, 0])

    lyr1 = layer.PressureLayer(100000.0, 60000.0)  # max at top of layer
    lyr2 = layer.PressureLayer(110000.0, 5000.0)  # out of bounds recovery
    lyr3 = layer.PressureLayer(100000.0, 10000.0)  # full layer

    assert (layer.layer_mean(lyr1, pres, data) == 1.25)
    assert (layer.layer_mean(lyr2, pres, data) == pytest.approx(1.111111))
    assert (layer.layer_mean(lyr3, pres, data) == pytest.approx(1.111111))



import test_accessor_ext as t


def test_01_str_attr_inplace_mutation():
    """
    Tests that a C++ expression like obj.attr("foo") += ... 
    can actually modify the object in-place.
    """
    a = t.test_str_attr_accessor_inplace_mutation()
    assert a.value == 1


def test_02_str_item_inplace_mutation():
    """
    Similar to test 01, but tests obj["foo"] (keyed attribute access)
    on the C++ side.
    """
    d = t.test_str_item_accessor_inplace_mutation()
    assert d.keys() == {"a"}
    assert d["a"] == 1


def test_03_num_item_list_inplace_mutation():
    """
    Similar to test 01, but tests l[n] (index access)
    on the C++ side, where l is an ``nb::list``.
    """
    l = t.test_num_item_list_accessor_inplace_mutation()
    assert len(l) == 1
    assert l[0] == 1


def test_04_obj_item_inplace_mutation():
    """
    Similar to test 01, but tests obj[h] (handle access)
    on the C++ side.
    """
    d = t.test_obj_item_accessor_inplace_mutation()
    assert len(d) == 1
    assert d.keys() == {0}
    assert d[0] == 1  # dict lookup



from typing import Annotated, overload

import numpy
from numpy.typing import NDArray


class HeightLayer:
    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, bottom: float, top: float, delta: float = 100.0) -> None: ...

    @property
    def bottom(self) -> float:
        """The bottom of the HeightLayer (meters)"""

    @bottom.setter
    def bottom(self, arg: float, /) -> None: ...

    @property
    def top(self) -> float:
        """The top of the HeightLayer (meters)"""

    @top.setter
    def top(self, arg: float, /) -> None: ...

    @property
    def delta(self) -> float:
        """The HeightLayer delta (increment) to use if iterating (meters)."""

    @delta.setter
    def delta(self, arg: float, /) -> None: ...

    def __repr__(self) -> str: ...

class PressureLayer:
    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, bottom: float, top: float, delta: float = -1000.0) -> None: ...

    @property
    def bottom(self) -> float:
        """The bottom of the PressureLayer (Pa)"""

    @bottom.setter
    def bottom(self, arg: float, /) -> None: ...

    @property
    def top(self) -> float:
        """The top of the PressureLayer (Pa)"""

    @top.setter
    def top(self, arg: float, /) -> None: ...

    @property
    def delta(self) -> float:
        """The PressureLayer delta (increment) to use if iterating (Pa)"""

    @delta.setter
    def delta(self, arg: float, /) -> None: ...

    def __repr__(self) -> str: ...

class LayerIndex:
    def __init__(self, kbot: int, ktop: int) -> None: ...

    @property
    def kbot(self) -> int:
        """
        The bottom index of a layer on a coordinate array (pressure or height).
        """

    @kbot.setter
    def kbot(self, arg: int, /) -> None: ...

    @property
    def ktop(self) -> int:
        """The top index of a layer on a coordinate array (pressure or height)."""

    @ktop.setter
    def ktop(self, arg: int, /) -> None: ...

    def __repr__(self) -> str: ...

@overload
def get_layer_index(layer: PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> LayerIndex:
    """
    Finds the interior array indices encapsulated by the given PressureLayer. 
    More specifically, the returned LayerIndex excludes the exact top and bottom 
    values (in mathematical notation, [bottom, top]). This behavior is due to the 
    fact many algorithms use interpolation to get the exact values of arbitrary
    top/bottom locations within a profile. 

    If layer.bottom or layer.top are out of bounds, this function 
    will truncate the layer to the coordinate range of data provided 
    by coord[] in an attempt to gracefully continue and produce a result. 
    This will modify the value of the given PressureLayer.

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressures

    Returns
    -------
    nwsspc.sharp.calc.layer.LayerIndex
        A LayerIndex with {kbot, ktop}.
    """

@overload
def get_layer_index(layer: HeightLayer, height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> LayerIndex:
    """
    Finds the interior array indices encapsulated by the given HeightLayer. 
    More specifically, the returned LayerIndex excludes the exact top and bottom 
    values (in mathematical notation, [bottom, top]). This behavior is due to the 
    fact many algorithms use interpolation to get the exact values of arbitrary
    top/bottom locations within a profile. 

    If layer.bottom or layer.top are out of bounds, this function 
    will truncate the layer to the coordinate range of data provided 
    by coord[] in an attempt to gracefully continue and produce a result. 
    This will modify the value of the given HeightLayer.

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.HeightLayer
    height : numpy.ndarray[dtype=float32]
        1D NumPy array of heights

    Returns
    -------
    nwsspc.sharp.calc.layer.LayerIndex
        A LayerIndex with {kbot, ktop}.
    """

def height_layer_to_pressure(layer: HeightLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], isAGL: bool = False) -> PressureLayer:
    """
    Converts a HeightLayer to a PressureLayer via interpolation, with the
    optional argument to convert the HeightLayer from meters AGL to MSL by adding
    the station height to the HeightLayer. If for some strange reason 
    you provide a HeightLayer that is out of the bounds of height[], then
    the bottom and top of the output layer will be set to MISSING.

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.HeightLayer
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure
    height : numpy.ndarray[dtype=float32]
        1D NumPy array of heights
    isAGL : bool 
        Whether or not the station height needs to be added for interpolation (default: False) 

    Returns
    -------
    nwsspc.sharp.calc.layer.PressureLayer
    """

def pressure_layer_to_height(layer: PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], toAGL: bool = False) -> HeightLayer:
    """
    Converts a PressureLayer to a HeightLayer via interpolation, with the 
    optional argument to convert the HeightLayer to meters AGL by subtracting off 
    the station height from the returned HeightLayer. If for some strange reason
    you provide a PressureLayer that is out of the bounds of pressure[], then 
    the bottom and top of the output layer will be set to MISSING. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure 
    height : numpy.ndarray[dtype=float32]
        1D NumPy array of heights
    toAGL : bool
        Whether or not to subtract the station height from the HeightLayer (default: False)

    Returns 
    -------
    nwsspc.sharp.calc.layer.HeightLayer
    """

@overload
def layer_min(layer: HeightLayer, height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], data: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[float, float]:
    """
    Returns the minimum value of the data array within the given HeightLayer. The 
    function bounds checks the layer by calling get_layer_index. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.HeightLayer 
    height : numpy.ndarray[dtype=float32]
        1D NumPy array of heights 
    data : numpy.ndarray[dtype=float32]
        1D array of data 

    Returns
    -------
    tuple[float, float]
        (min_value, level_of_min)
    """

@overload
def layer_min(layer: PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], data: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[float, float]:
    """
    Returns the minimum value of the data array within the given PressureLayer. The 
    function bounds checks the layer by calling get_layer_index. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer 
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of pressure 
    data : numpy.ndarray[dtype=float32] 
        1D array of data 

    Returns
    -------
    tuple[float, float]
        (min_value, level_of_min)
    """

@overload
def layer_max(layer: HeightLayer, height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], data: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[float, float]:
    """
    Returns the maximum value observed within the given data array over
    the given HeightLayer. The function bounds checks the layer by calling 
    get_layer_index. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.HeightLayer 
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values 
    data : numpy.ndarray[dtype=float32] 
        1D NumPy array of data values

    Returns
    -------
    tuple[float, float]
        (max_value, level_of_max)
    """

@overload
def layer_max(layer: PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], data: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[float, float]:
    """
    Returns the maximum value observed within the given data array over the given 
    PressureLayer. The function bounds checks the layer by calling get_layer_index. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer 
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure values 
    data : numpy.ndarray[dtype=float32] 
        1D NumPy array of data values

    Returns
    -------
    tuple[float, float]
        (max_value, level_of_max)
    """

@overload
def layer_mean(PressureLayer: PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], data: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Computes the pressure-weighted mean value of a field over 
    a given PressureLayer. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer 
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of pressure values 
    data : numpy.ndarray[dtype=float32] 
        1D NumPy array of data values

    Returns
    -------
    float
    """

@overload
def layer_mean(HeightLayer: HeightLayer, height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], data: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], isAGL: bool = False) -> float:
    """
    Computes the pressure-weighted mean value of a field over 
    a given HeightLayer. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.HeightLayer 
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure values 
    data : numpy.ndarray[dtype=float32]
        1D NumPy array of data values
    isAGL : bool
        Whether or not the surface station height should be added to the HeightLayer (default: False)

    Returns
    -------
    float
    """

@overload
def integrate_layer_trapz(layer: HeightLayer, data: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], integ_sign: int = 0, weighted: bool = False) -> float:
    """
    Returns a trapezoidal integration of the given data array over 
    the given HeightLayer. There is an additional argument that 
    determines whether this is a weighted average or not. The sign 
    of the integration may be passed as well, i.e. integrating only 
    positive or negative area, by passing a 1, 0, or -1 to integ_sign. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.HeightLayer 
    data : numpy.ndarray[dtype=float32] 
        1D NumPy array of data values 
    height : numpy.ndarray[dtype=float32]
        1D NumPy array of height values 
    integ_sign : int 
        The sign of the area to integrate (-1: negative; 1: positive; 0: both; default: 0)
    weighted : bool 
        Boolean determining whether or not the integration is weighted by the coordinate array 

    Returns
    -------
    float
    """

@overload
def integrate_layer_trapz(layer: PressureLayer, data: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], integ_sign: int = 0, weighted: bool = False) -> float:
    """
    Returns the trapezoidal integration of the given data array over 
    the given PressureLayer. There is an additional argument that 
    determines whether this is a weighted average or not. The sign 
    of the integration may be passed as well, i.e. integrating only 
    positive or negative area, by passing a 1, 0, or -1 to integ_sign. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer
    data : numpy.ndarray[dtype=float32] 
        1D NumPy array of data values
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of pressure values 
    integ_sign : int 
        The sign of the area to integrate (-1: negative; 1: positive; 0: both; default: 0)
    weighted : bool 
        Boolean determining whether or not the integration is weighted by the coordinate array

    Returns 
    -------
    float
    """



#!/usr/bin/env python3
#

import os
import re

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_copybutton",
    "sphinxcontrib.rsvgconverter",
    "sphinxcontrib.moderncmakedomain",
    "sphinx.ext.intersphinx",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = [".templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "nanobind"
copyright = "2023, Wenzel Jakob"
author = "Wenzel Jakob"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# Read the listed version
VERSION_REGEX = re.compile(
    r"^\s*#\s*define\s+NB_VERSION_([A-Z]+)\s+(.*)$", re.MULTILINE)

this_directory = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))

with open(os.path.join(this_directory, "include/nanobind/nanobind.h")) as f:
    matches = dict(VERSION_REGEX.findall(f.read()))
    version = "{MAJOR}.{MINOR}.{PATCH}".format(**matches)

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = [".build", "release.rst"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = "any"

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = 'monokai'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "furo"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<version> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#  html_static_path = ["_static"]

#  html_css_files = [ ]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'h', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'r', 'sv', 'tr'
# html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
# html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
# html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = "nanobind_doc"

# -- Options for LaTeX output ---------------------------------------------

latex_engine = "pdflatex"

latex_elements = {
    'papersize': 'a4paper',
    'pointsize': '10pt',
    "classoptions": ",openany,oneside",
    "preamble": r"""
\usepackage{MnSymbol}
\DeclareUnicodeCharacter{25CB}{\ensuremath{\circ}}
\DeclareUnicodeCharacter{25CF}{\ensuremath{\bullet}}
\DeclareUnicodeCharacter{21B5}{\ensuremath{\rhookswarrow}}
\DeclareUnicodeCharacter{2194}{\ensuremath{\leftrightarrow}}
""",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "nanobind.tex", "nanobind Documentation", author, "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = 'nanobind-logo.png'

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True

primary_domain = "cpp"
highlight_language = "cpp"



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import nwsspc.sharp.calc as sharplib 
import datetime
import breathe
import sys, os 

sys.path.insert(0, os.path.abspath("../"))

version_tuple = sharplib.__version_tuple__
version = sharplib.__version__
year = datetime.datetime.now(datetime.UTC).year
doc_version = os.environ.get('DOC_VERSION', 'dev' if 'dev' in version else version).lstrip('v')

project = 'SHARPlib'
copyright = f'2024-{year}, NOAA/NWS/NCEP Storm Prediction Center'
author = 'Kelton Halbert'
release = f'{version_tuple[0]}.{version_tuple[1]}.{version_tuple[2]}'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.graphviz',
    'sphinx.ext.intersphinx',
    'sphinx.ext.githubpages',
    'breathe',
]

html_static_path = ['_static']
templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
add_function_parentheses = False

autoclass_content = "both"
autodoc_member_order = "bysource"
autodoc_docstring_signature = True
autodoc_typehints = "both"
autodo_type_aliases = {
    "PressureLayer": "nwsspc.sharp.calc.layer.PressureLayer",
    "HeightLayer": "nwsspc.sharp.calc.layer.HeightLayer",
    "LayerIndex": "nwsspc.sharp.calc.layer.LayerIndex",
}
autodoc_typehints_format = "short"
autodoc_preserve_defauls = True

napoleon_numpy_docstring = True 
napoleon_google_docstring = False
napoleon_preprocess_types = True

breathe_projects = {
    "SHARPlib": "doxyxml/"
}
breathe_domain_by_extension = {"h": "cpp"}

breathe_projects_source = {
    "SHARPlib" : (
        "../", [
            "include/SHARPlib/constants.h", 
            "include/SHARPlib/interp.h",
            "include/SHARPlib/layer.h",
            "include/SHARPlib/parcel.h",
            "include/SHARPlib/thermo.h",
            "include/SHARPlib/winds.h",
            "include/SHARPlib/params/convective.h",
            "include/SHARPlib/params/fire.h",
            "include/SHARPlib/params/winter.h",
            "src/SHARPlib/interp.cpp",
            "src/SHARPlib/layer.cpp",
            "src/SHARPlib/parcel.cpp",
            "src/SHARPlib/thermo.cpp",
            "src/SHARPlib/winds.cpp",
            "src/SHARPlib/params/convective.cpp",
            "src/SHARPlib/params/fire.cpp",
            "src/SHARPlib/params/winter.cpp",
        ] 
    )
}

breathe_default_project = "SHARPlib"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"

html_theme_options = {
    "show_toc_level": 2,
    "logo": {
        "text": f"SHARPlib",
        "image_light": "_static/logo.png",
        "image_dark": "_static/logo.png",
    },
    "collapse_navigation": True,
    'external_links': [
        {'name': 'Examples', 'url': 'https://github.com/keltonhalbert/SHARPlib/tree/main/examples/Python'},
        {'name': 'Release Notes', 'url': 'https://github.com/keltonhalbert/SHARPlib/releases'},
    ],
    'icon_links': [
        {
            'name': 'GitHub',
            'url': 'https://github.com/keltonhalbert/SHARPlib',
            'icon': 'fa-brands fa-github',
            'type': 'fontawesome',
        },
        {
            'name': 'Bluesky',
            'url': 'https://bsky.app/profile/stormscale.io',
            'icon': 'fa-brands fa-bluesky',
            'type': 'fontawesome',
        },
    ],
    'use_edit_page_button': False,
    'navbar_align': 'left',
    'navbar_start': ['navbar-logo', 'version-switcher'],
    'navbar_center': ['navbar-nav'],
    'header_links_before_dropdown': 6,
    'navbar_persistent': ['search-button'],
    'navbar_end': ['navbar-icon-links', 'theme-switcher'],
    'switcher': {
        'json_url': 'https://keltonhalbert.github.io/SHARPlib/versions.json',
        'version_match': doc_version
    },
}


html_title = f"{project} v{version} Manual"
html_last_updated_fmt = '%b %d, %Y'
html_context = {"default_mode": "dark"}
html_use_modindex = True
html_copy_source = False
html_domain_indices = False
html_file_suffix = '.html'


intersphinx_mapping = {
    'numpy': ('https://numpy.org/doc/stable/', None),
}



import contextlib
import os
import platform
import re
import shutil
import sys
from pathlib import Path
from typing import Any, Generator

import setuptools
from setuptools.command import build_ext

IS_WINDOWS = platform.system() == "Windows"
IS_MAC = platform.system() == "Darwin"
IS_LINUX = platform.system() == "Linux"

# hardcoded SABI-related options. Requires that each Python interpreter
# (hermetic or not) participating is of the same major-minor version.
py_limited_api = sys.version_info >= (3, 12)
options = {"bdist_wheel": {"py_limited_api": "cp312"}} if py_limited_api else {}


def is_cibuildwheel() -> bool:
    return os.getenv("CIBUILDWHEEL") is not None


@contextlib.contextmanager
def _maybe_patch_toolchains() -> Generator[None, None, None]:
    """
    Patch rules_python toolchains to ignore root user error
    when run in a Docker container on Linux in cibuildwheel.
    """

    def fmt_toolchain_args(matchobj):
        suffix = "ignore_root_user_error = True"
        callargs = matchobj.group(1)
        # toolchain def is broken over multiple lines
        if callargs.endswith("\n"):
            callargs = callargs + "    " + suffix + ",\n"
        # toolchain def is on one line.
        else:
            callargs = callargs + ", " + suffix
        return "python.toolchain(" + callargs + ")"

    CIBW_LINUX = is_cibuildwheel() and IS_LINUX
    module_bazel = Path("MODULE.bazel")
    content: str = module_bazel.read_text()
    try:
        if CIBW_LINUX:
            module_bazel.write_text(
                re.sub(
                    r"python.toolchain\(([\w\"\s,.=]*)\)",
                    fmt_toolchain_args,
                    content,
                )
            )
        yield
    finally:
        if CIBW_LINUX:
            module_bazel.write_text(content)


class BazelExtension(setuptools.Extension):
    """A C/C++ extension that is defined as a Bazel BUILD target."""

    def __init__(self, name: str, bazel_target: str, **kwargs: Any):
        super().__init__(name=name, sources=[], **kwargs)

        self.bazel_target = bazel_target
        stripped_target = bazel_target.split("//")[-1]
        self.relpath, self.target_name = stripped_target.split(":")


class BuildBazelExtension(build_ext.build_ext):
    """A command that runs Bazel to build a C/C++ extension."""

    def run(self):
        for ext in self.extensions:
            self.bazel_build(ext)
        super().run()
        # explicitly call `bazel shutdown` for graceful exit
        self.spawn(["bazel", "shutdown"])

    def copy_extensions_to_source(self):
        """
        Copy generated extensions into the source tree.
        This is done in the ``bazel_build`` method, so it's not necessary to
        do again in the `build_ext` base class.
        """
        pass

    def bazel_build(self, ext: BazelExtension) -> None:
        """Runs the bazel build to create the package."""
        temp_path = Path(self.build_temp)

        # We round to the minor version, which makes rules_python
        # look up the latest available patch version internally.
        python_version = "{0}.{1}".format(*sys.version_info[:2])

        bazel_argv = [
            "bazel",
            "run",
            ext.bazel_target,
            f"--symlink_prefix={temp_path / 'bazel-'}",
            f"--compilation_mode={'dbg' if self.debug else 'opt'}",
            # C++17 is required by nanobind
            f"--cxxopt={'/std:c++17' if IS_WINDOWS else '-std=c++17'}",
            f"--@rules_python//python/config_settings:python_version={python_version}",
        ]

        if ext.py_limited_api:
            bazel_argv += ["--@nanobind_bazel//:py-limited-api=cp312"]

        if IS_WINDOWS:
            # Link with python*.lib.
            for library_dir in self.library_dirs:
                bazel_argv.append("--linkopt=/LIBPATH:" + library_dir)
        elif IS_MAC:
            # C++17 needs macOS 10.14 at minimum
            bazel_argv.append("--macos_minimum_os=10.14")

        with _maybe_patch_toolchains():
            self.spawn(bazel_argv)

        if IS_WINDOWS:
            suffix = ".pyd"
        else:
            suffix = ".abi3.so" if ext.py_limited_api else ".so"

        # copy the Bazel build artifacts into setuptools' libdir,
        # from where the wheel is built.
        pkgname = "google_benchmark"
        pythonroot = Path("bindings") / "python" / "google_benchmark"
        srcdir = temp_path / "bazel-bin" / pythonroot
        libdir = Path(self.build_lib) / pkgname
        for root, dirs, files in os.walk(srcdir, topdown=True):
            # exclude runfiles directories and children.
            dirs[:] = [d for d in dirs if "runfiles" not in d]

            for f in files:
                fp = Path(f)
                should_copy = False
                # we do not want the bare .so file included
                # when building for ABI3, so we require a
                # full and exact match on the file extension.
                if "".join(fp.suffixes) == suffix:
                    should_copy = True
                elif fp.suffix == ".pyi":
                    should_copy = True
                elif Path(root) == srcdir and f == "py.typed":
                    # copy py.typed, but only at the package root.
                    should_copy = True

                if should_copy:
                    shutil.copyfile(root / fp, libdir / fp)


setuptools.setup(
    cmdclass=dict(build_ext=BuildBazelExtension),
    package_data={"google_benchmark": ["py.typed", "*.pyi"]},
    ext_modules=[
        BazelExtension(
            name="google_benchmark._benchmark",
            bazel_target="//bindings/python/google_benchmark:benchmark_stubgen",
            py_limited_api=py_limited_api,
        )
    ],
    options=options,
)



from typing import Annotated, overload

import numpy
from numpy.typing import NDArray

import nwsspc.sharp.calc.layer
import nwsspc.sharp.calc.parcel
import nwsspc.sharp.calc.winds


@overload
def effective_inflow_layer(lifter: nwsspc.sharp.calc.parcel.lifter_wobus, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtemp: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], cape_thresh: float = 100.0, cinh_thresh: float = -250.0, mupcl: nwsspc.sharp.calc.parcel.Parcel | None = None) -> nwsspc.sharp.calc.layer.PressureLayer:
    """
    Computes the Effective Inflow Layer, or the layer of the atmosphere
    beliefed to be the primary source of inflow for supercell thunderstorms. 
    The Effective Inflow Layer, and its use in computing shear and storm 
    relative helicity, is described by Thompson et al. 2007: 
    https://www.spc.noaa.gov/publications/thompson/effective.pdf

    Standard/default values for cape_thresh and cinh_thresh have been 
    experimentally determined to be cape_thresh = 100 J/kg and 
    cinh_thresh = -250.0 J/kg. If an empty parcel object is passed via the 
    'mupcl' kwarg, the Most Unstable parcel found during the EIL search will 
    be returned. 

    References 
    ----------
    Thompson et al. 2007: https://www.spc.noaa.gov/publications/thompson/effective.pdf

    Parameters 
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_wobus 
    pressure : numpy.ndarray[dtype=float32]
        A 1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of height values (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of temperature values (K)
    dewpoint : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of dewpoint values (K)
    virtemp : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of virtual temperature values (K)
    cape_thresh : float, default = 100.0
        The CAPE threshold used to compute the Effective Inflow Layer 
    cinh_thresh : float, default = -250.0 
        The CINH threshold used to compute the Effective Inflow Layer
    mupcl : None or nwsspc.sharp.calc.parcel.Parcel, optional

    Returns 
    -------
    nwsspc.sharp.calc.layer.PressureLayer
        The Effective Inflow Layer
    """

@overload
def effective_inflow_layer(lifter: nwsspc.sharp.calc.parcel.lifter_cm1, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtemp: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], cape_thresh: float = 100.0, cinh_thresh: float = -250.0, mupcl: nwsspc.sharp.calc.parcel.Parcel | None = None) -> nwsspc.sharp.calc.layer.PressureLayer:
    """
    Computes the Effective Inflow Layer, or the layer of the atmosphere
    beliefed to be the primary source of inflow for supercell thunderstorms. 
    The Effective Inflow Layer, and its use in computing shear and storm 
    relative helicity, is described by Thompson et al. 2007:
    https://www.spc.noaa.gov/publications/thompson/effective.pdf

    Standard/default values for cape_thresh and cinh_thresh have been 
    experimentally determined to be cape_thresh = 100 J/kg and 
    cinh_thresh = -250.0 J/kg. If an empty parcel object is passed via the 
    'mupcl' kwarg, the Most Unstable parcel found during the EIL search will 
    be returned. 

    References 
    ----------
    Thompson et al. 2007: https://www.spc.noaa.gov/publications/thompson/effective.pdf

    Parameters 
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_cm1 
    pressure : numpy.ndarray[dtype=float32]
        A 1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of height values (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of temperature values (K)
    dewpoint : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of dewpoint values (K)
    virtemp : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of virtual temperature values (K)
    cape_thresh : float, default = 100.0
        The CAPE threshold used to compute the Effective Inflow Layer 
    cinh_thresh : float, default = -250.0 
        The CINH threshold used to compute the Effective Inflow Layer
    mupcl : None or nwsspc.sharp.calc.parcel.Parcel, optional

    Returns 
    -------
    nwsspc.sharp.calc.layer.PressureLayer
        The Effective Inflow Layer
    """

@overload
def storm_motion_bunkers(pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], mean_wind_layer_agl: nwsspc.sharp.calc.layer.HeightLayer, wind_shear_layer_agl: nwsspc.sharp.calc.layer.HeightLayer, leftMover: bool = False, pressureWeighted: bool = False) -> nwsspc.sharp.calc.winds.WindComponents:
    """
    Estimates the supercell storm motion using the Bunkers et al. 2000 method 
    described in the following paper:
    https://doi.org/10.1175/1520-0434(2000)015%3C0061:PSMUAN%3E2.0.CO;2

    This does not use any of the updated methods described by Bunkers et al. 2014, 
    which uses Effective Inflow Layer metrics to get better estimates of storm 
    motion, especially when considering elevated convection. 

    References 
    ----------

    Buners et al. 2000: https://doi.org/10.1175/1520-0434(2000)015%3C0061:PSMUAN%3E2.0.CO;2

    Parameters 
    ----------
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values (meters)
    u_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of U wind component values (m/s)
    v_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of V wind compnent values (m/s)
    mean_wind_layer_agl : nwsspc.sharp.calc.layer.HeightLayer 
        HeightLayer (AGL) for computing the mean wind 
    wind_shear_layer_agl : nwsspc.sharp.calc.layer.HeightLayer 
        HeightLayer (AGL) for computing wind_shear
    leftMover : bool 
        Whether to compute left mover supercell motion (default: False)
    pressureWeighted : bool 
        Whether to use the pressure weighted mean wind (default: False)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindComponents
        U, V wind components of storm motion (m/s)
    """

@overload
def storm_motion_bunkers(pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], eff_infl_lyr: nwsspc.sharp.calc.layer.PressureLayer, mupcl: nwsspc.sharp.calc.parcel.Parcel, leftMover: bool = False) -> nwsspc.sharp.calc.winds.WindComponents:
    """
    Estimates supercell storm motion using the Bunkers et al. 2014 
    method described in the following paper:
    http://dx.doi.org/10.15191/nwajom.2014.0211

    This method is parcel based, using a mean-wind vector defined as the 
    pressure-weighted mean wind between the Effective Inflow Layer surface 
    (see effective_inflow_layer routine) and 65% of the depth between that 
    surface and the most unstable parcel's Equilibrium Level. This method 
    produces the same storm motion estimate for surface based supercells, 
    and captures the motion of elevated supercells better than the 
    Bunkers 2000 method. 

    The input parameters of eff_infl_lyr and mupcl (effective inflow layer 
    pressure bounds and the most unstable parcel, respectively) are required
    to be precomputed and passed to this routine. These are expensive 
    operations that are presumed to be computed at some other point 
    in the analysis pipeline. 

    References
    ----------
    Bunkers et al. 2014: http://dx.doi.org/10.15191/nwajom.2014.0211

    Parameters 
    ----------
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32]
        1D NumPy array of height values (meters)
    u_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of U wind component values (m/s)
    v_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of V wind component values (m/s)
    eff_infl_lyr : nwsspc.sharp.calc.layer.PressureLayer 
        Effective Inflow Layer PressureLayer
    mupcl : nwsspc.sharp.calc.parcel.Parcel 
        Most Unstable Parcel 
    leftMover : bool 
        Whether or not to compute left moving supercell motion (default: False)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindComponents
        U, V wind components of storm motion (m/s)
    """

def mcs_motion_corfidi(pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[nwsspc.sharp.calc.winds.WindComponents, nwsspc.sharp.calc.winds.WindComponents]:
    """
    Compute the Corfidi upshear and downshear MCS motion vectors.

    Estimates the mesoscale convective system (MCS) motion vectors for upshear 
    and downshear propagating convective systems as in Corfidi et al. 2003.
    The method is based on observations that MCS motion is a function of 
    1) the advection of existing cells by the mean wind and 
    2) the propagation of new convection relative to existing storms.

    References
    ----------
    Corfidi et al. 2003: https://www.spc.noaa.gov/publications/corfidi/mcs2003.pdf

    Parameters 
    ----------
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32]
        1D NumPy array of height values (meters)
    u_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of u-wind components (m/s)
    v_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of v-wind components (m/s)

    Returns 
    -------
    tuple[nwsspc.sharp.calc.winds.WindComponents, nwsspc.sharp.calc.winds.WindComponents]
        (upshear, downshear)
    """

def effective_bulk_wind_difference(pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], effective_inflow_layer: nwsspc.sharp.calc.layer.PressureLayer, equilibrium_level_pressure: float) -> nwsspc.sharp.calc.winds.WindComponents:
    """
    Compute the Effective Bulk Wind Difference 

    The effective bulk wind difference is the wind shear between 
    the bottom height of the effective inflow layer, and 50% of 
    the equilibrium level depth. This is analogous to the usage 
    of 0-6 km wind shear, but allows more flexibility for elevated 
    convection. Returns MISSING if the effective inflow layer or 
    equilibrium level pressure are MISSING.
    """

def energy_helicity_index(cape: float, helicity: float) -> float:
    """
    Computes the Energy Helicity Index.

    EHI is a composite parameter based on the premise that 
    storm rotation shoudl be maximized when CAPE is large 
    and SRH is large. Typically, the layers used for helicity 
    are either 0-1 km AGL or 0-3 km AGL.

    References 
    ----------
    https://doi.org/10.1175/1520-0434(2003)18%3C530:RSATFP%3E2.0.CO;2

    Parameters 
    ----------
    CAPE : float 
        Convective Available Potential Energy (J/kg)
    helicity : float 
        Storm Relative Helicity (m^2 / s^2 a.k.a J/kg)

    Returns
    -------
    float
        Energy Helicity Index (umitless)
    """

def significant_tornado_parameter(pcl: nwsspc.sharp.calc.parcel.Parcel, lcl_hght_agl: float, storm_relative_helicity: float, bulk_wind_difference: float) -> float:
    """
    Computes the Significant Tornado Parameter.


    The Significant Tornado Parameter is used to diagnose environments
    where tornadoes are favored. STP traditionally comes in two flavors:
    fixed-layer, and effective-layer. Fixed-layer STP expects surface-based
    CAPE, the surface-based LCL, 0-1 km storm-relative helicity, the
    0-6 km bulk wind difference, and the surface-based CINH. For the
    effective inflow layer based STP, use 100mb mixed-layer CAPE,
    100mb mixed-layer LCL height AGL, effective-layer srh, the
    effective layer bulk wind difference, and the 100mb mixed-layer
    CINH. NOTE: The effective bulk wind difference is the shear between
    the bottom of the effective inflow layer and 50% of the height of the
    equilibrium level of the most unstable parcel.

    References
    ----------
    Thompson et al 2012: https://www.spc.noaa.gov/publications/thompson/waf-env.pdf

    Parameters 
    ----------
    pcl : nwsspc.sharp.calc.parcel.Parcel 
        For effective-layer STP, a mixed-layer parcel, and for fixed-layer STP, a surface-based parcel 
    lcl_hght_agl : float 
        The parcel LCL height in meters
    storm_relative_helicity : float 
        For effective-layer STP, effective SRH, and for fixed-layer, 0-1 km SRH (m^2 / s^2)
    bulk_wind_difference : float 
        For effective-layer STP, effective BWD, and for fixed-layer STP, 0-6 km BWD (m/s)

    Returns
    -------
    float 
        The Significant Tornado Parameter
    """

def supercell_composite_parameter(mu_cape: float, eff_srh: float, eff_shear: float) -> float:
    """
    Computes the Supercell Composite Parameter. 

    The supercell composite parameter is used to diagnose environments
    where supercells are favored. Requires computing most unstable
    CAPE, effective layer storm relative helicity, and effective
    bulk shear. Effective bulk shear is the vector difference between
    the winds at the bottom of the effective inflow layer, and 50% of
    the equilibrium level height. It is similar to the 0-6 km shear
    vector, but allows for elevated supercell thunderstorms.

    The left-moving supercell composite parameter can be computed by
    providing effective SRH calculated using the bunkers left-moving
    storm motion, and will return negative values.

    References
    ----------
    Thompson et al 2003: https://www.spc.noaa.gov/publications/thompson/ruc_waf.pdf

    Thompson et al 2007: https://www.spc.noaa.gov/publications/thompson/effective.pdf

    Thompson et al 2012: https://www.spc.noaa.gov/publications/thompson/waf-env.pdf



    Parameters 
    ----------
    mu_cape : float 
        The CAPE of the Most Unstable Parcel (J/kg)
    eff_srh : float 
        Effective inflow layer Storm Relative Helicity (m^2/s^2) 
    eff_shear : float 
        Effective layer shear (m/s)

    Returns
    -------
    float
        Supercell Composite Parameter (unitless)
    """

def significant_hail_parameter(mu_pcl: nwsspc.sharp.calc.parcel.Parcel, lapse_rate_700_500mb: float, tmpk_500mb: float, freezing_level_agl: float, shear_0_6km: float) -> float:
    """
    Compute the significant hail parameter, given a precomputed most-unstable parcel, 
    the 700-500 mb lapse rate, the 500mb temperature, the height (AGL) of the 
    freezing level, and the 0-6 km shear magnitude.

    The Sig. Hail Parameter (SHIP) was developed using a large database of 
    surface-modified, observed severe hail proximity soundings. It is based on 
    parameters, and is meant to delineate between SIG (>=2" diameter) and NON-SIG
    (<2" diameter) hail environments.

    SHIP = [(MUCAPE j/kg) * (Mixing Ratio of MU PARCEL g/kg) *  
            (700-500mb LAPSE RATE c/km) * (-500mb TEMP C) *
            (0-6km Shear m/s) ] / 42,000,000

    0-6 km shear is confined to a range of 7-27 m s-1, mixing ratio is confined to 
    a range of 11-13.6 g kg-1, and the 500 mb temperature is set to -5.5 C for 
    any warmer values.

    Once the initial version of SHIP is calculated, the values are modified in 
    the following scenarios:

    1) If MUCAPE < 1300 J kg-1, SHIP = SHIP * (MUCAPE/1300); 2) if 700-500 mb 
    lapse rate < 5.8 C km-1, SHIP = SHIP * (lr75/5.8); 3) if freezing 
    level < 2400 m AGL, SHIP = SHIP * (fzl/2400)

    It is important to note that SHIP is NOT a forecast hail size.

    Since SHIP is based on the RAP depiction of MUCAPE - unrepresentative MUCAPE 
    "bullseyes" may cause a similar increase in SHIP values. This typically occurs 
    when bad surface observations get into the RAP model.

    Developed in the same vein as the STP and SCP parameters, values of SHIP 
    greater than 1.00 indicate a favorable environment for SIG hail. Values greater 
    than 4 are considered very high. In practice, maximum contour values of 1.5-2.0 
    or higher will typically be present when SIG hail is going to be reported. 

    Parameters 
    ----------
    mu_pcl : nwsspc.sharp.calc.parcel.Parcel 
        A precomputed Most Unstable parcel
    lapse_rate_700_500mb : float
        The 700-500 mb lapse rate (K/km)
    tmpk_500mb : float 
        The 500mb temperature (K)
    freezing_level_agl : float 
        The height of the freezing level (AGL, meters)
    shear_0_6km : float
        The 0-6 km shear vector magnitude (m/s)

    Returns
    -------
    float
        The significant hail parameter
    """

def derecho_composite_parameter(dcape: float, mucape: float, shear_0_6km: float, mean_wind_0_6km: float) -> float:
    """
    The Derecho Composite Parameter (DCP) is based on a dataset of 113 derecho 
    events compiled by the Evans and Boswell (2001) study. It is intended to 
    hightlight environments favorable for cold pool/outflow driven convective 
    events. The physical mechanisms behind this parameter focus on cold pool 
    production (DCAPE), ability to sustain strong convection (MUCAPE), 
    convective organization (0 - 6 km shear), and sufficient deep-layer flow 
    within the environment.

    References
    ----------
    Evans and Doswell 2001: https://doi.org/10.1175/1520-0434(2001)016%3C0329:EODEUP%3E2.0.CO;2

    Parameters
    ----------
    dcape : float 
        Downdraft Convective Available Potential Energy (J/kg)
    mucape : float 
        Most Unstable Parcel Convective Available Potential Energy (J/kg)
    shear_0_6km : float 
        Shear magnitude in the 0 - 6 km layer AGL (m/s)
    mean_wind_0_6km : float 
        Mean wind magnitude in the 0 - 6 km layer AGL (m/s)

    Returns
    -------
    float 
        The Derecho Composite Parameter
    """

def large_hail_parameter(mu_pcl: nwsspc.sharp.calc.parcel.Parcel, lapse_rate_700_500mb: float, hail_growth_zone: nwsspc.sharp.calc.layer.PressureLayer, storm_motion: nwsspc.sharp.calc.winds.WindComponents, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Computes the Large Hail Parameter (LHP). The LHP is a multi-ingredient, 
    composite index that includes thermodynamics and kinematics to attempt 
    to detect environments that support very large hail. LHP has shown skill 
    when differentiationg environments that support hail >= 3.5 in from those 
    with < 2.0 in.

    References
    ----------
    Johnson and Sugden 2014: https://ejssm.org/archives/wp-content/uploads/2021/09/vol9-5.pdf

    Parameters 
    ----------
    mu_pcl : nwsspc.sharp.calc.parcel.Parcel 
        A previously computed most unstable parcel with CAPE and an equilibrium level
    lapse_rate_700_500mb : float 
        700 - 500 hPa Lapse Rate (K)
    hail_growth_zone : nwsspc.sharp.calc.layer.PressureLayer
        The layer of the atmosphere encompasing the Hail Growth Zone (Pa)
    storm_motion : nwsspc.sharp.calc.winds.WindComponents 
        The storm motion vector to be used (m/s)
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32]
        1D NumPy array of height values (m)
    u_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of u_wind values (m/s)
    v_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of v_wind values (m/s)
    """

@overload
def convective_temperature(lifter: nwsspc.sharp.calc.parcel.lifter_wobus, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtual_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], mixratio: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], cinh_thresh: float = -1.0) -> float:
    """
    Computes the convective temperature by iteratively lifting parcels from 
    the surface using a lowest 100 hPa mean mixing ratio and increasing 
    surface temperatures to find the first parcel that reaches the CINH 
    threshold. The first guess is the current surface temperature.

    Parameters 
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_wobus 
    pressure : numpy.ndarray[dtype=float32]
        A 1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of height values (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of temperature values (K)
    virtemp : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of virtual temperature values (K)
    mixratio : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of water vapor mixing ratio values (g/g)
    cinh_thresh : float, default = -1.0 
        The CINH threshold used to compute the convective temperature (J/kg)

    Returns
    -------
    float
        The convective temperature (K)
    """

@overload
def convective_temperature(lifter: nwsspc.sharp.calc.parcel.lifter_cm1, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtual_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], mixratio: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], cinh_thresh: float = -1.0) -> float:
    """
    Computes the convective temperature by iteratively lifting parcels from 
    the surface using a lowest 100 hPa mean mixing ratio and increasing 
    surface temperatures to find the first parcel that reaches the CINH 
    threshold. The first guess is the current surface temperature.

    Parameters 
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_cm1 
    pressure : numpy.ndarray[dtype=float32]
        A 1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of height values (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of temperature values (K)
    virtemp : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of virtual temperature values (K)
    mixratio : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of water vapor mixing ratio values (g/g)
    cinh_thresh : float, default = -1.0 
        The CINH threshold used to compute the convective temperature (J/kg)

    Returns
    -------
    float
        The convective temperature (K)
    """

def precipitable_water(layer: nwsspc.sharp.calc.layer.PressureLayer, pres: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], mixr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Given a PressureLayer to integrate over, compute the precipitable water 
    from the given pressure and mixing ratio arrays.

    Parameters 
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer 
        a PressureLayer over which to integrate (Pa)
    pres : numpy.ndarray[dtype=float32] 
        1D NumPy array of presssure values (Pa)
    mixr : numpy.ndarray[dtype=float32] 
        1D NumPy array of water vapor mixing ratio values (unitless)

    Returns
    -------
    float
        Precipitable water content (mm)
    """

def hail_growth_layer(pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> nwsspc.sharp.calc.layer.PressureLayer:
    """
    Performs a top-down search for the hail growth zone, 
    returning the PressureLayer bounds. This zone is defined 
    as the pressure layer that contains the portion of the 
    profile between -30C and -10C.

    Parameters 
    ----------
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure values (Pa)
    temperature : numpy.ndarray[dtype=float32]
        1D NumPy array of temperature values (K)

    Returns 
    -------
    nwsspc.sharp.calc.layer.PressureLayer
        The PressureLayer containing the hail growth zone
    """

def dendritic_layer(pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> nwsspc.sharp.calc.layer.PressureLayer:
    """
    Performs a top-down search for the dendritic growth zone, 
    returning the PressureLayer bounds. This zone is defined 
    as the pressure layer that contains the portion of the 
    profile between -17C and -12C.

    Parameters 
    ----------
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of pressure values (Pa)
    temperature : numpy.ndarray[dtype=float32]
        1D NumPy array of temperature values (K)

    Returns 
    -------
    nwsspc.sharp.calc.layer.PressureLayer
        The PressureLayer containing the dendritic growth zone
    """

def snow_squall_parameter(wetbulb_2m: float, mean_relh_0_2km: float, delta_thetae_0_2km: float, mean_wind_0_2km: float) -> float:
    """
    The Snow Squall Parameter is a non-dimensional parameter that combines 
    several ingredients believed to be beneficial for identifying snow squall 
    environments by identifying the overlap of low-level potential instability, 
    sufficient moisture, and strong low-level winds.


    References
    ----------
    Banacos et al. 2014: https://www.weather.gov/media/btv/research/Snow%20Squalls%20Forecasting%20and%20Hazard%20Mitigation.pdf

    Parameters 
    ----------
    wetbulb_2m : float 
        The surface wetbulb temperature, used to mask the parameter (K)
    mean_relh_0_2km : float 
        The mean relative humidity between the surface and 2 km AGL (fraction)
    delta_thetae_0_2km : float 
        The difference in equivalent potential temperature between 2 km AGL and the surface (K)
    mean_wind_0_2km : float 
        The mean wind speed between the surface and 2 km AGL (m/s)

    Returns 
    -------
    float 
        The Snow Squall Parameter
    """

@overload
def equilibrium_moisture_content(temperature: float, rel_humidity: float) -> float:
    """
    Compute the equilibrium moisture content for fuel 
    as in Simard (1968).

    Parameters
    ----------
    temperature : float 
        The air temperature (K)
    rel_humidity : float 
        Relative Humidity (fraction)

    Returns 
    -------
    float 
        Equilibrium Moisture Content (fraction)
    """

@overload
def equilibrium_moisture_content(temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], rel_humidity: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the equilibrium moisture content for fuel 
    as in Simard (1968).

    Parameters
    ----------
    temperature : numpy.ndarray[dtype=float32] 
        The air temperature (K)
    rel_humidity : numpy.ndarray[dtype=float32]
        Relative Humidity (fraction)

    Returns 
    -------
    numpy.ndarray[dtype=float32]
        Equilibrium Moisture Content (fraction)
    """

@overload
def fosberg_fire_index(temperature: float, rel_humidity: float, wind_speed: float) -> float:
    """
    Compute the Fosberg Fire-Weather Index (FWWI) as in Fosberg (1978).

    Parameters 
    ----------
    temperature : float 
        The air temperature (K)
    rel_humidity : float 
        The relative humidity (fraction)
    wind_speed : float 
        Wind speed (m/s)

    Returns 
    -------
    float 
        Fosberg Fire-Weather Index
    """

@overload
def fosberg_fire_index(temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], rel_humidity: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], wind_speed: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the Fosberg Fire-Weather Index (FWWI) as in Fosberg (1978).

    Parameters 
    ----------
    temperature : numpy.ndarray[dtype=float32]
        The air temperature (K)
    rel_humidity : numpy.ndarray[dtype=float32]
        The relative humidity (fraction)
    wind_speed : numpy.ndarray[dtype=float32]
        Wind speed (m/s)

    Returns 
    -------
    numpy.ndarray[dtype=float32]
        Fosberg Fire-Weather Index
    """

@overload
def pyrocumulonimbus_firepower_threshold(lifter: nwsspc.sharp.calc.parcel.lifter_cm1, mix_layer: nwsspc.sharp.calc.layer.PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], mixratio: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtual_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], potential_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], pcl: nwsspc.sharp.calc.parcel.Parcel | None = None, phi: float = 6.67e-05, beta_incr: float = 0.005) -> float:
    """
    Computes the Pyrocumulonimbus Firepower Threshold (PFT), or the minimum 
    amount of firepower required to generate pyrocumulonimbus clouds for a 
    given atmospheric profile. Requires a PressureLayer to define a mixing 
    layer used to average values of potential temperature, mixing ratio, 
    and wind speed. The beta increment determines how to vary the plume 
    buoyancy factor, with smaller values resulting in more iteration steps. 
    Phi is the fire moisture to potential temperature increment ratio.

    Default values for beta_incr and phi are 0.005 and 6.67e-5, respectively.
    If a parcel is passed, the values will be set with the PFT fire parcel.

    References 
    ----------
    Tory et al. 2018: https://journals.ametsoc.org/view/journals/mwre/146/8/mwr-d-17-0377.1.xml
    Tory et al. 2021: https://journals.ametsoc.org/view/journals/wefo/36/2/WAF-D-20-0027.1.xml

    Parameters 
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_cm1
    mix_layer : nwsspc.sharp.calc.layer.PressureLayer
        A mixing layer to take average values of the environment from
    pressure : numpy.ndarray[dtype=float32]
        A 1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of height values (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of temperature values (K)
    mixratio : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of water vapor mixing ratio values (g/g)
    virtual_temperature : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of virtual temperature values (K)
    u_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of U wind component values (m/s)
    v_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of V wind compnent values (m/s)
    potential_temperature : numpy.ndarray[dtype=float32]
        1D NumPy array of potential temperature values (K)
    pcl : None or nwsspc.sharp.calc.parcel.Parcel, optional
        If a parcel is provided, returns the parcel for the PFT
    phi : float 
        The fire moisture to potential temperature increment ratio (kg/(kg*K))
    beta_incr : float 
        The fire plume buoyancy factor (unitless)

    Returns 
    -------
    float 
        The PyroCB Firepower Threshold (Watts)
    """

@overload
def pyrocumulonimbus_firepower_threshold(lifter: nwsspc.sharp.calc.parcel.lifter_wobus, mix_layer: nwsspc.sharp.calc.layer.PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], mixratio: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtual_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], potential_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], pcl: nwsspc.sharp.calc.parcel.Parcel | None = None, phi: float = 6.67e-05, beta_incr: float = 0.005) -> float:
    """
    Computes the Pyrocumulonimbus Firepower Threshold (PFT), or the minimum 
    amount of firepower required to generate pyrocumulonimbus clouds for a 
    given atmospheric profile. Requires a PressureLayer to define a mixing 
    layer used to average values of potential temperature, mixing ratio, 
    and wind speed. The beta increment determines how to vary the plume 
    buoyancy factor, with smaller values resulting in more iteration steps. 
    Phi is the fire moisture to potential temperature increment ratio.

    Default values for beta_incr and phi are 0.005 and 6.67e-5, respectively.
    If a parcel is passed, the values will be set with the PFT fire parcel.

    References 
    ----------
    Tory et al. 2018: https://journals.ametsoc.org/view/journals/mwre/146/8/mwr-d-17-0377.1.xml
    Tory et al. 2021: https://journals.ametsoc.org/view/journals/wefo/36/2/WAF-D-20-0027.1.xml

    Parameters 
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_wobus
    mix_layer : nwsspc.sharp.calc.layer.PressureLayer
        A mixing layer to take average values of the environment from
    pressure : numpy.ndarray[dtype=float32]
        A 1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of height values (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of temperature values (K)
    mixratio : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of water vapor mixing ratio values (g/g)
    virtual_temperature : numpy.ndarray[dtype=float32] 
        A 1D NumPy array of virtual temperature values (K)
    u_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of U wind component values (m/s)
    v_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of V wind compnent values (m/s)
    potential_temperature : numpy.ndarray[dtype=float32]
        1D NumPy array of potential temperature values (K)
    pcl : None or nwsspc.sharp.calc.parcel.Parcel, optional
        If a parcel is provided, returns the parcel for the PFT
    phi : float 
        The fire moisture to potential temperature increment ratio (kg/(kg*K))
    beta_incr : float 
        The fire plume buoyancy factor (unitless)

    Returns 
    -------
    float 
        The PyroCB Firepower Threshold (Watts)
    """



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python benchmarking utilities.

Example usage:
  import google_benchmark as benchmark

  @benchmark.register
  def my_benchmark(state):
      ...  # Code executed outside `while` loop is not timed.

      while state:
        ...  # Code executed within `while` loop is timed.

  if __name__ == '__main__':
    benchmark.main()
"""

import atexit

from absl import app

from google_benchmark import _benchmark
from google_benchmark._benchmark import (
    Counter as Counter,
    State as State,
    kMicrosecond as kMicrosecond,
    kMillisecond as kMillisecond,
    kNanosecond as kNanosecond,
    kSecond as kSecond,
    o1 as o1,
    oAuto as oAuto,
    oLambda as oLambda,
    oLogN as oLogN,
    oN as oN,
    oNCubed as oNCubed,
    oNLogN as oNLogN,
    oNone as oNone,
    oNSquared as oNSquared,
)

__version__ = "1.9.1"


class __OptionMaker:
    """A stateless class to collect benchmark options.

    Collect all decorator calls like @option.range(start=0, limit=1<<5).
    """

    class Options:
        """Pure data class to store options calls, along with the benchmarked function."""

        def __init__(self, func):
            self.func = func
            self.builder_calls = []

    @classmethod
    def make(cls, func_or_options):
        """Make Options from Options or the benchmarked function."""
        if isinstance(func_or_options, cls.Options):
            return func_or_options
        return cls.Options(func_or_options)

    def __getattr__(self, builder_name):
        """Append option call in the Options."""

        # The function that get returned on @option.range(start=0, limit=1<<5).
        def __builder_method(*args, **kwargs):
            # The decorator that get called, either with the benchmared function
            # or the previous Options
            def __decorator(func_or_options):
                options = self.make(func_or_options)
                options.builder_calls.append((builder_name, args, kwargs))
                # The decorator returns Options so it is not technically a decorator
                # and needs a final call to @register
                return options

            return __decorator

        return __builder_method


# Alias for nicer API.
# We have to instantiate an object, even if stateless, to be able to use __getattr__
# on option.range
option = __OptionMaker()


def register(undefined=None, *, name=None):
    """Register function for benchmarking."""
    if undefined is None:
        # Decorator is called without parenthesis so we return a decorator
        return lambda f: register(f, name=name)

    # We have either the function to benchmark (simple case) or an instance of Options
    # (@option._ case).
    options = __OptionMaker.make(undefined)

    if name is None:
        name = options.func.__name__

    # We register the benchmark and reproduce all the @option._ calls onto the
    # benchmark builder pattern
    benchmark = _benchmark.RegisterBenchmark(name, options.func)
    for name, args, kwargs in options.builder_calls[::-1]:
        getattr(benchmark, name)(*args, **kwargs)

    # return the benchmarked function because the decorator does not modify it
    return options.func


def _flags_parser(argv):
    argv = _benchmark.Initialize(argv)
    return app.parse_flags_with_usage(argv)


def _run_benchmarks(argv):
    if len(argv) > 1:
        raise app.UsageError("Too many command-line arguments.")
    return _benchmark.RunSpecifiedBenchmarks()


def main(argv=None):
    return app.run(_run_benchmarks, argv=argv, flags_parser=_flags_parser)


# Methods for use with custom main function.
initialize = _benchmark.Initialize
run_benchmarks = _benchmark.RunSpecifiedBenchmarks
atexit.register(_benchmark.ClearRegisteredBenchmarks)
